"""
The inventory module defines the ``Inventory`` and ``InventoryHP`` classes. Instances of these
classes contain one or more nuclides each with an associated amount (number of atoms). The decay
of the nuclide(s) can be calculated by using the ``decay()`` method. The ``Inventory`` class
performs calculations using double-precision floats. The ``InventoryHP`` class performs
calculations using SymPy high numerical precision operations. A ``DecayData`` dataset is associated
with ``Inventory`` and ``InventoryHP`` instances (default is rd.DEFAULTDATA).

The docstring code examples assume that ``radioactivedecay`` has been imported
as ``rd``:

.. highlight:: python
.. code-block:: python

    >>> import radioactivedecay as rd

"""

import numbers
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import matplotlib
import numpy as np
from scipy import sparse
from sympy import Integer, Matrix, exp, nsimplify
from sympy.core.expr import Expr

from radioactivedecay.converters import (
    QuantityConverter,
    QuantityConverterFloat,
    QuantityConverterSympy,
    UnitConverter,
    UnitConverterFloat,
    UnitConverterSympy,
)
from radioactivedecay.decaydata import (
    DEFAULTDATA,
    DecayData,
    DecayMatrices,
    DecayMatricesScipy,
    DecayMatricesSympy,
)
from radioactivedecay.nuclide import Nuclide
from radioactivedecay.plots import decay_graph
from radioactivedecay.utils import (
    add_dictionaries,
    parse_nuclide,
    sort_dictionary_alphabetically,
    sort_list_according_to_dataset,
)

# functools.singledispatchmethod is new in Python version 3.8
# this block adds support manually for Python <3.8
# the except block can be removed once min required Python version is 3.8
try:
    from functools import singledispatchmethod
except ImportError:
    from functools import singledispatch, update_wrapper

    def singledispatchmethod(func: Callable[..., Any]):  # type:ignore
        """Adds singledispatch support for instance methods of a class."""

        dispatcher = singledispatch(func)

        def wrapper(*args, **kwargs):  # type:ignore
            return dispatcher.dispatch(args[1].__class__)(*args, **kwargs)

        wrapper.register = dispatcher.register
        update_wrapper(wrapper, func)
        return wrapper


# pylint: disable=too-many-arguments, too-many-lines, too-many-locals


class AbstractInventory(ABC):
    """
    Template class for float (SciPy) or SymPy Inventory classes. Each AbstractInventory instance
    stores nuclides and associated quantities (numbers of atoms) in the contents dictionary. A
    decay dataset is associated with each instance.

    Parameters
    ----------
    contents : dict
        Dictionary containing nuclide strings/canonical ids or Nuclide instances as keys and
        quantities (activities, number of atoms, masses or moles) as values.
    units : str, optional
        Units of the contents dictionary values. Specify either 'num' for number of atoms, or an
        activity, mass or moles unit. e.g. 'Bq', 'kBq', 'Ci', 'g', 'kg', 'mol', 'kmol'. Default is
        'Bq'.
    check : bool, optional
        Check for the validity of the contents dictionary (nuclides are in the decay dataset,
        and quantities provided are physical, etc.). Default is True.
    decay_data : DecayData, optional
        Decay dataset (default is the ICRP-107 dataset).

    Attributes
    ----------
    contents : dict
        Dictionary containing nuclide strings as keys and number of atoms of each nuclide as
        values. Nuclides are sorted alphabetically in this dictionary.
    decay_data : DecayData
        Decay dataset.
    decay_matrices : DecayMatrices
        SciPy or SymPy version of a DecayMatrices instance associated with the decay dataset.

    """

    def __init__(
        self,
        contents: Dict[Union[str, int, Nuclide], Union[float, Expr]],
        units: str = "Bq",
        check: bool = True,
        decay_data: DecayData = DEFAULTDATA,
    ) -> None:
        self.decay_data = decay_data
        self.decay_matrices = self._get_decay_matrices()

        if check is True:
            contents_with_parsed_keys: Dict[
                str, Union[float, Expr]
            ] = self._parse_nuclides(
                contents, self.decay_data.nuclides, self.decay_data.dataset_name
            )
            self._check_values(contents_with_parsed_keys)
        else:
            contents_with_parsed_keys = contents
        contents_sorted = sort_dictionary_alphabetically(contents_with_parsed_keys)
        self.contents = self._convert_to_number(
            contents_sorted, units, self.decay_data.dataset_name
        )

    @staticmethod
    def _parse_nuclides(
        contents: Dict[Union[str, int, Nuclide], Union[float, Expr]],
        nuclides: np.ndarray,
        dataset_name: str,
    ) -> Dict[str, Union[float, Expr]]:
        """
        Checks that nuclide keys in the contents dictionary. Converts Nuclide instances into
        nuclide name strings. Converts nuclide name strings and ids into Ab-XY format.
        """

        return {
            (
                parse_nuclide(nuc.nuclide, nuclides, dataset_name)
                if isinstance(nuc, Nuclide)
                else parse_nuclide(nuc, nuclides, dataset_name)
            ): inp
            for nuc, inp in contents.items()
        }

    @staticmethod
    def _check_values(contents: Dict[str, Union[float, Expr]]) -> None:
        """Checks that the nuclide quantities in contents (dictionary values) are valid."""

        for nuc, inp in contents.items():
            if isinstance(inp, (numbers.Number, Expr)):
                if inp >= 0:
                    continue
            raise ValueError(f"{inp} is not a valid quantity of nuclide {nuc}.")

    def _get_atomic_mass(self, nuc: str) -> Union[float, Expr]:
        """Returns the appropriate atomic mass."""

        return self.decay_matrices.atomic_masses[self.decay_data.nuclide_dict[nuc]]

    def _get_decay_const(self, nuc: str) -> Union[float, Expr]:
        """Returns the appropriate atomic mass."""

        return self.decay_matrices.decay_consts[self.decay_data.nuclide_dict[nuc]]

    @abstractmethod
    def _get_decay_matrices(self) -> DecayMatrices:
        """Returns the appropriate DecayMatrices instance."""

    @staticmethod
    @abstractmethod
    def _get_quantity_converter() -> Type[QuantityConverter]:
        """Returns the appropriate QuantityConverter instance."""

    @abstractmethod
    def _get_unit_converter(self) -> Type[UnitConverter]:
        """Returns the appropriate UnitConverter instance."""

    @abstractmethod
    def _get_year_conv(self) -> Union[float, Expr]:
        """Returns the appropriate number of days in a year."""

    def _convert_to_number(
        self,
        contents: Dict[str, Union[float, Expr]],
        units: str,
        dataset_name: str,
    ) -> Dict[str, Union[float, Expr]]:
        """
        Converts an inventory dictionary where the values are masses, moles or activities to one
        where the values are number of atoms.

        Parameters
        ----------
        contents : dict
            Dictionary containing nuclide strings or Nuclide objects as keys, and masses,
            moles, or activities as values.
        units : str
            Units of the values in the input dictionary.

        Returns
        -------
        dict
            Inventory dictionary where the values are the number of atoms of the nuclide.

        Raises
        ------
        ValueError
            If the units supplied are invalid or an activity is supplied for a stable nuclide.

        """

        if units == "num":
            contents_as_numbers = contents
        elif units in self._get_unit_converter().activity_units:
            contents_as_numbers = {}
            for nuc, act in contents.items():
                decay_const = self._get_decay_const(nuc)
                if decay_const == 0:
                    raise ValueError(
                        f"{nuc} is a stable radionuclide in {dataset_name} decay dataset. "
                        "Initializing an inventory with an activity for a stable nuclide (i.e. "
                        f"{act} {units} for {nuc}) has no physical meaning. Initialize using a "
                        "mass or mole quantity instead."
                    )
                contents_as_numbers[
                    nuc
                ] = self._get_quantity_converter().activity_to_number(
                    self._get_unit_converter().activity_unit_conv(act, units, "Bq"),
                    decay_const,
                )
        elif units in self._get_unit_converter().moles_units:
            contents_as_numbers = {
                nuc: self._get_quantity_converter().moles_to_number(
                    self._get_unit_converter().moles_unit_conv(mol, units, "mol")
                )
                for nuc, mol in contents.items()
            }
        elif units in self._get_unit_converter().mass_units:
            contents_as_numbers = {
                nuc: self._get_quantity_converter().mass_to_number(
                    self._get_unit_converter().mass_unit_conv(mass, units, "g"),
                    self._get_atomic_mass(nuc),
                )
                for nuc, mass in contents.items()
            }
        else:
            raise ValueError(f"{units} is not a supported unit.")

        return contents_as_numbers

    @property
    def nuclides(self) -> List[str]:
        """
        Returns a list of the nuclides in the inventory.

        Examples
        --------
        >>> rd.Inventory({'Tc-99m': 2.3, 'I-123': 5.8}, 'Bq').nuclides
        ['I-123', 'Tc-99m']

        """

        return list(self.contents.keys())

    def numbers(self) -> Dict[str, float]:
        """
        Returns a dictionary containing the number of atoms of each nuclide within the inventory.

        Examples
        --------
        >>> rd.Inventory({'Tc-99m': 2.3, 'I-123': 5.8}, 'Bq').numbers()
        {'I-123': 399738.47946141585, 'Tc-99m': 71852.27235544211}

        """

        return self.contents

    def activities(self, units: str = "Bq") -> Dict[str, float]:
        """
        Returns a dictionary containing the activity of each nuclide within the inventory.

        Parameters
        ----------
        units : str, optional
            Activity units for output, e.g. 'Bq', 'kBq', 'mBq', 'Ci', 'dpm'...
            Deafult is 'Bq'.

        Examples
        --------
        >>> rd.Inventory({'Tc-99m': 2300, 'I-123': 5800}, 'Bq').activities('kBq')
        {'I-123': 5.8, 'Tc-99m': 2.3}

        """

        activities = {
            nuc: self._get_unit_converter().activity_unit_conv(
                self._get_quantity_converter().number_to_activity(
                    num, self._get_decay_const(nuc)
                ),
                "Bq",
                units,
            )
            for nuc, num in self.contents.items()
        }

        return activities

    def masses(self, units: str = "g") -> Dict[str, float]:
        """
        Returns a dictionary containing the mass of each nuclide within the inventory

        Parameters
        ----------
        units : str, optional
            Mass units for output, e.g. 'Bq', 'g', 'kg', 'mg', 'ton'...
            Deafult is 'g'.

        Examples
        --------
        >>> rd.Inventory({'Tc-99m': 2.3, 'I-123': 5.8}, 'Bq').masses('pg')
        {'I-123': 8.158243973887584e-05, 'Tc-99m': 1.1800869622748502e-05}

        """

        masses = {
            nuc: self._get_unit_converter().mass_unit_conv(
                self._get_quantity_converter().number_to_mass(
                    num, self._get_atomic_mass(nuc)
                ),
                "g",
                units,
            )
            for nuc, num in self.contents.items()
        }

        return masses

    def moles(self, units: str = "mol") -> Dict[str, float]:
        """
        Returns a dictionary containing the number of atoms of each nuclide within the inventory in
        moles.

        Parameters
        ----------
        units : str, optional
            Moles units, e.g. 'mmol', 'mol', 'kmol'...
            Deafult is 'mol'.

        Examples
        --------
        >>> rd.Inventory({'Tc-99m': 2.3, 'I-123': 5.8}, 'Bq').moles()
        {'I-123': 6.637813617983513e-19, 'Tc-99m': 1.1931350531142702e-19}

        """

        moles = {
            nuc: self._get_unit_converter().moles_unit_conv(
                self._get_quantity_converter().number_to_moles(num), "mol", units
            )
            for nuc, num in self.contents.items()
        }

        return moles

    def activity_fractions(self) -> Dict[str, float]:
        """
        Returns a dictionary containing the activity fraction of each nuclide within the inventory.

        Examples
        --------
        >>> rd.Inventory({'Tc-99m': 2.3, 'I-123': 5.8}, 'Bq').activity_fractions()
        {'I-123': 0.7160493827160493, 'Tc-99m': 0.2839506172839506}

        """

        total_activity = sum(self.activities().values())
        activity_fractions = {
            nuc: act / total_activity for nuc, act in self.activities().items()
        }

        return activity_fractions

    def mass_fractions(self) -> Dict[str, float]:
        """
        Returns a dictionary containing the mass fraction of each nuclide within the inventory.

        Examples
        --------
        >>> rd.Inventory({'Tc-99m': 2.3, 'I-123': 5.8}, 'Bq').mass_fractions()
        {'I-123': 0.8736297770616593, 'Tc-99m': 0.12637022293834066}

        """

        total_mass = sum(self.masses().values())
        mass_fractions = {nuc: mass / total_mass for nuc, mass in self.masses().items()}

        return mass_fractions

    def mole_fractions(self) -> Dict[str, float]:
        """
        Returns a dictionary containing the mole fraction of each nuclide within the inventory.

        Examples
        --------
        >>> rd.Inventory({'Tc-99m': 2.3, 'I-123': 5.8}, 'Bq').mole_fractions()
        {'I-123': 0.8476385041932588, 'Tc-99m': 0.15236149580674116}

        """

        total_number = sum(self.numbers().values())
        mole_fractions = {
            nuc: num / total_number for nuc, num in self.numbers().items()
        }

        return mole_fractions

    def __len__(self) -> int:
        """
        Returns number of nuclides in the inventory.
        """

        return len(self.contents)

    def add(
        self,
        add_contents: Dict[Union[str, int, Nuclide], float],
        units: str = "Bq",
    ) -> None:
        """
        Adds a dictionary of nuclides and associated quantities (numbers/activities/masses/moles)
        to the inventory.

        Parameters
        ----------
        add_contents : dict
            Dictionary containing nuclide strings, canonical ids or Nuclide objects as keys and the
            amount of each nuclide (with specified units) as values which are added to the
            Inventory.
        units : str, optional
            Units of the values in the dictionary (e.g. 'Bq', 'Ci', 'g', 'mol', 'num'; 'Bq' is the
            default).

        Examples
        --------
        >>> inv = rd.Inventory({'H-3': 1.0}, 'Bq')
        >>> inv.add({'C-14': 2.0}, 'kBq')
        >>> inv.add({190400000: 20.0})
        >>> inv.activities()
        {'K-40': 20.0, 'C-14': 2000.0, 'H-3': 1.0}

        """

        other = Inventory(add_contents, units, True, self.decay_data)
        self.contents = (self + other).contents

    def subtract(
        self,
        sub_contents: Dict[Union[str, int, Nuclide], float],
        units: str = "Bq",
    ) -> None:
        """
        Subtracts a dictionary of nuclides and associated numbers/activities/masses
        from the inventory.

        Parameters
        ----------
        sub_contents : dict
            Dictionary containing nuclide strings or Nuclide objects as keys
            and the amount of each nuclide (with specified units) as values
            which are subtracted from the Inventory.
        units : str, optional
            Units of the values in the dictionary (e.g. 'Bq', 'Ci', 'g', 'mol', 'num'; 'Bq' is the
            default).

        Examples
        --------
        >>> inv = rd.Inventory({'K-40': 20.0, 'C-14': 2.0, 'H-3': 1.0}, 'Bq')
        >>> inv.subtract({'H-3': 1.0})
        >>> inv.subtract({190400000: 20.0})
        >>> inv.activities()
        {'C-14': 2.0, 'H-3': 0.0}

        """

        other = Inventory(sub_contents, units, True, self.decay_data)
        self.contents = (self - other).contents

    def __add__(self, other: "AbstractInventory") -> "AbstractInventory":
        """Defines + operator to add two Inventory objects together."""

        if self.decay_data != other.decay_data:
            raise ValueError(
                f"Decay datasets do not match. {self.decay_data.dataset_name} and "
                f"{other.decay_data.dataset_name}"
            )
        new_contents = add_dictionaries(self.contents, other.contents)
        return self.__class__(new_contents, "num", False, self.decay_data)

    def __sub__(self, other: "AbstractInventory") -> "AbstractInventory":
        """Defines - operator to subtract one inventory object from another."""

        if self.decay_data != other.decay_data:
            raise ValueError(
                f"Decay datasets do not match. {self.decay_data.dataset_name} and "
                f"{other.decay_data.dataset_name}"
            )
        sub_contents = other.contents.copy()
        sub_contents.update(
            (nuclide, number * -1.0) for nuclide, number in sub_contents.items()
        )
        new_contents = add_dictionaries(self.contents, sub_contents)
        return self.__class__(new_contents, "num", False, self.decay_data)

    def __mul__(self, const: Union[float, Expr]) -> "AbstractInventory":
        """
        Defines * operator to multiply all quantities of nuclides in an inventory by a constant.
        """

        new_contents = self.contents.copy()
        for nuclide, number in new_contents.items():
            new_contents[nuclide] = number * const

        return self.__class__(new_contents, "num", False, self.decay_data)

    def __rmul__(self, const: Union[float, Expr]) -> "AbstractInventory":
        """
        Defines * operator to multiply all quantities of nuclides in an Inventory by a constant.
        """

        return self.__mul__(const)

    def __truediv__(self, const: Union[float, Expr]) -> "AbstractInventory":
        """
        Defines / operator to divide all quantities of nuclides in an inventory by a constant.
        """
        new_contents = self.contents.copy()
        for nuclide, number in new_contents.items():
            new_contents[nuclide] = number / const

        return self.__class__(new_contents, "num", False, self.decay_data)

    @singledispatchmethod
    def remove(
        self, delete: Union[str, int, Nuclide, List[Union[str, int, Nuclide]]]
    ) -> None:
        """
        Removes nuclide(s) from the inventory.

        Parameters
        ----------
        delete : str or int or Nuclide or list
            Nuclide string, canonical id, Nuclide object or list of nuclide strings, canonical ids
            or Nuclide objects to delete from the Inventory object.

        Examples
        --------
        >>> inv = rd.Inventory({'Be-10': 2.0, 'C-14': 3.0, 'H-3': 1.0, 'K-40': 4.0}, 'Bq')
        >>> inv.remove('H-3')
        >>> inv.activities()
        {'Be-10': 2.0, 'C-14': 3.0, 'K-40': 4.0}
        >>> inv.remove(['Be-10', 'K-40'])
        >>> inv.activities()
        {'C-14': 3.0}

        """
        raise NotImplementedError(
            "Method takes a nuclide string, a canonical id, a Nuclide instance, or list of nuclide"
            " strings, canonical ids or Nuclide instances as a parameter."
        )

    @remove.register(str)
    def _(
        self, delete: str
    ) -> Callable[[Dict[str, float], str, bool, DecayData], None]:
        """Remove nuclide string from this inventory."""

        delete = parse_nuclide(
            delete, self.decay_data.nuclides, self.decay_data.dataset_name
        )
        new_contents = self.contents.copy()
        if delete not in new_contents:
            raise ValueError(f"{delete} does not exist in this inventory.")
        new_contents.pop(delete)

        self.contents = new_contents

    @remove.register(int)
    def _(
        self, delete: int
    ) -> Callable[[Dict[str, float], str, bool, DecayData], None]:
        """Remove nuclide string from this inventory."""

        delete_str = parse_nuclide(
            delete, self.decay_data.nuclides, self.decay_data.dataset_name
        )
        new_contents = self.contents.copy()
        if delete_str not in new_contents:
            raise ValueError(f"{delete_str} does not exist in this inventory.")
        new_contents.pop(delete_str)

        self.contents = new_contents

    @remove.register(Nuclide)
    def _(
        self, delete: Nuclide
    ) -> Callable[[Dict[str, float], str, bool, DecayData], None]:
        """Remove Nuclide object from this inventory."""

        delete_str = parse_nuclide(
            delete.nuclide, self.decay_data.nuclides, self.decay_data.dataset_name
        )
        new_contents = self.contents.copy()
        if delete_str not in new_contents:
            raise ValueError(f"{delete_str} does not exist in this inventory.")
        new_contents.pop(delete_str)

        self.contents = new_contents

    @remove.register(list)
    def _(
        self, delete: List[Union[str, int, Nuclide]]
    ) -> Callable[[Dict[str, float], str, bool, DecayData], None]:
        """Remove list of nuclide(s) from this inventory."""

        delete_list = [
            parse_nuclide(
                nuc.nuclide, self.decay_data.nuclides, self.decay_data.dataset_name
            )
            if isinstance(nuc, Nuclide)
            else parse_nuclide(
                nuc, self.decay_data.nuclides, self.decay_data.dataset_name
            )
            for nuc in delete
        ]
        new_contents = self.contents.copy()
        for nuc in delete_list:
            if nuc not in new_contents:
                raise ValueError(f"{nuc} does not exist in this inventory.")
            new_contents.pop(nuc)

        self.contents = new_contents

    def _convert_decay_time(
        self, decay_time: Union[float, Expr], units: str
    ) -> Union[float, Expr]:
        """
        Converts a decay time period into seconds.
        """

        return self._get_unit_converter().time_unit_conv(
            decay_time, units, "s", self._get_year_conv()
        )

    def _setup_decay_calc(
        self,
    ) -> Tuple[Union[np.ndarray, Matrix], List[int], Union[sparse.csr_matrix, Matrix]]:
        """
        Setup variables for a decay calculation.
        """

        vector_n0 = self.decay_matrices.vector_n0.copy()
        indices_set = set()
        for nuclide in self.contents:
            idx = self.decay_data.nuclide_dict[nuclide]
            vector_n0[idx] = self.contents[nuclide]
            indices_set.update(self.decay_data.scipy_data.matrix_c[:, idx].nonzero()[0])
        indices = list(indices_set)

        matrix_e = self.decay_matrices.matrix_e.copy()

        return vector_n0, indices, matrix_e

    def _perform_decay_calc(
        self,
        vector_n0: Union[np.ndarray, Matrix],
        matrix_e: Union[sparse.csr_matrix, Matrix],
    ) -> Union[np.ndarray, Matrix]:
        """
        Perform decay calculation matrix multiplication.
        """

        vector_nt = (
            self.decay_matrices.matrix_c
            @ matrix_e
            @ self.decay_matrices.matrix_c_inv
            @ vector_n0
        )
        return vector_nt

    @abstractmethod
    def decay(self, decay_time: float, units: str = "s") -> "AbstractInventory":
        """
        Returns a new inventory calculated from the radioactive decay of the current inventory for
        decay_time.
        """

    @abstractmethod
    def cumulative_decays(
        self, decay_time: float, units: str = "s"
    ) -> Dict[str, float]:
        """
        Calculates the total number of decays of each nuclide in the inventory between t=0 and
        t=decay_time. Note no results are reported for stable nuclides, as cumulative decays is
        zero.
        """

    def half_lives(self, units: str = "s") -> Dict[str, Union[float, str]]:
        """
        Returns dictionary of half-lives of the nuclides in the inventory in your chosen time
        units, or as a human-readable string with appropriate units.

        Parameters
        ----------
        units : str, optional
            Units for half-life. Options are 'ps', 'ns', 'μs', 'us', 'ms', 's', 'm', 'h', 'd', 'y',
            'ky', 'My', 'By', 'Gy', 'Ty', 'Py', and common spelling variations. Default is 's', i.e.
            seconds. Use 'readable' to get strings of the half-lives in human-readable units.

        Returns
        -------
        dict
            Dictionary with nuclide strings as keys and half-life floats or human-readable
            half-life strings as values.

        Examples
        --------
        >>> inv = rd.Inventory({'C-14': 1.0, 'K-40': 2.0}, 'Bq')
        >>> inv.half_lives('y')
        {'C-14': 5700.0, 'K-40': 1251000000.0}
        >>> inv.half_lives('readable')
        {'C-14': '5.70 ky', 'K-40': '1.251 By'}
        """

        return {nuc: self.decay_data.half_life(nuc, units) for nuc in self.contents}

    def progeny(self) -> Dict[str, List[str]]:
        """
        Returns dictionary with the direct progeny of the nuclides in the inventory.

        Returns
        -------
        dict
            Dictionary with nuclide strings as keys and lists of the direct progeny of each
            nuclide, ordered by decreasing branching fraction, as values.

        Examples
        --------
        >>> inv = rd.Inventory({'C-14': 1.0, 'K-40': 2.0}, 'Bq')
        >>> inv.progeny()
        {'C-14': ['N-14'], 'K-40': ['Ca-40', 'Ar-40'])

        """

        return {nuc: Nuclide(nuc, self.decay_data).progeny() for nuc in self.contents}

    def branching_fractions(self) -> Dict[str, List[float]]:
        """
        Returns dictionary with the branching fractions of the direct progeny of the nuclides
        in the inventory.

        Returns
        -------
        dict
            Dictionary with nuclide strings as keys and lists of the branching fractions to
            the direct progeny of each nuclide as values.

        Examples
        --------
        >>> inv = rd.Inventory({'C-14': 1.0, 'K-40': 2.0}, 'Bq')
        >>> inv.branching_fractions()
        {'C-14': [1.0], 'K-40': [0.8914, 0.1086])

        """

        return {
            nuc: Nuclide(nuc, self.decay_data).branching_fractions()
            for nuc in self.contents
        }

    def decay_modes(self) -> Dict[str, List[str]]:
        """
        Returns dictionary with the decay modes of the direct progeny of the nuclides in the
        inventory. Note: the decay mode strings returned are not lists of all the different
        radiation types emitted during the parent to progeny decay processes. They are the labels
        defined in the decay dataset to classify the decay type (e.g. '\u03b1', '\u03b2-' or 'IT').

        Returns
        -------
        dict
            Dictionary with nuclide strings as keys and lists of the decay modes of the parent
            nuclide

        Examples
        --------
        >>> inv = rd.Inventory({'C-14': 1.0, 'K-40': 2.0}, 'Bq')
        >>> inv.decay_modes()
        {'C-14': ['\u03b2-'], 'K-40': ['\u03b2-', '\u03b2+ \u0026 EC'])

        """

        return {
            nuc: Nuclide(nuc, self.decay_data).decay_modes() for nuc in self.contents
        }

    def plot(  # type: ignore
        self,
        xmax: float,
        xunits: str = "s",
        xmin: float = 0.0,
        xscale: str = "linear",
        yscale: str = "linear",
        ymin: float = 0.0,
        ymax: Optional[float] = None,
        yunits: str = "Bq",
        display: Union[str, List[str]] = "all",
        order: str = "dataset",
        npoints: int = 501,
        fig: Optional[matplotlib.figure.Figure] = None,
        axes: Optional[matplotlib.axes.Axes] = None,
        **kwargs,
    ) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
        """
        Plots a decay graph showing the change in activity of the inventory over time. Creates
        matplotlib fig, axes objects if they are not supplied. Returns fig, axes tuple.

        Parameters
        ----------
        xmax : float
            Maximum decay time on x-axis.
        xunits : str, optional
            Units for decay times (default is 's', i.e. seconds). Options are 'ps', 'ns', 'us',
            'ms', 's', 'm', 'h', 'd', 'y', 'ky', 'My', 'Gy', 'Ty', 'Py', and some of the common
            spelling variations of these time units.
        xmin : float, optional
            Minimum decay time on x-axis (default is 0.0 for linear x-axis, 0.1 for log x-axis).
        xscale : str, optional
            The time axis scale type to apply ('linear' or 'log', default is 'linear').
        yscale : str, optional
            The y-axis scale type to apply ('linear' or 'log', default is 'linear').
        ymin : float, optional
            Minimum value for the y-axis (default is 0.0 for ``yscale='linear'``, or 0.95x the
            minimum quantity that occurs over the decay period for ``yscale='log'``).
        ymax : None or float, optional
            Maximum value for the y-axis. Default is None, which sets the limit to 1.05x the
            maximum quantity that occurs over the decay period.
        yunits : str, optional
            Units to display on the y-axis e.g. 'Bq', 'kBq', 'Ci', 'g', 'mol', 'num',
            'activity_frac', 'mass_frac', 'mol_frac'. Default is 'Bq'.
        display : str or list, optional
            Only display the nuclides within this list on the graph. Use this parameter when
            you want to choose specific nuclide decay curves shown on the graph, either by
            supplying a string (to show one nuclide) or a list of strings (to show multiple).
            Default is 'all', which displays all nuclides present upon decay of the inventory.
        order : str, optional
            Order to display the nuclide decay curves on the graph if you do not specify the
            order via the display parameter. Default order is by 'dataset', which follows the order
            of the nuclides in the decay dataset (highest to lowest nuclides in the decay
            chains). Use 'alphabetical' if you want the nuclides to be ordered alphabetically.
        npoints : None or int, optional
            Number of time points used to plot graph (default is 501 for normal precision decay
            calculations, or 51 for high precision decay calculations).
        fig : None or matplotlib.figure.Figure, optional
            matplotlib figure object to use, or None makes ``radioactivedecay`` create one (default
            is None).
        axes : None or matplotlib.axes.Axes, optional
            matplotlib axes object to use, or None makes ``radioactivedecay`` create one (default
            is None).
        **kwargs, optional
            All additional keyword arguments to supply to matplotlib plot() function.

        Returns
        -------
        fig : matplotlib.figure.Figure
            matplotlib figure object used to plot decay chain.
        axes : matplotlib.axes.Axes
            matplotlib axes object used to plot decay chain.

        Raises
        ------
        ValueError
            If the order parameter is invalid.

        """

        if xscale == "linear":
            time_points = np.linspace(xmin, xmax, num=npoints)
        else:
            if xmin == 0.0:
                xmin = 0.1
            time_points = np.logspace(np.log10(xmin), np.log10(xmax), num=npoints)

        if display == "all":
            if order == "dataset":
                display = sort_list_according_to_dataset(
                    self.decay(0).nuclides, self.decay_data.nuclide_dict
                )
            elif order == "alphabetical":
                display = self.decay(0).nuclides
            else:
                raise ValueError(
                    f"{order} is not a valid string for the order parameter."
                )
        else:
            if isinstance(display, str):
                display = [display]
            display = [
                parse_nuclide(
                    rad, self.decay_data.nuclides, self.decay_data.dataset_name
                )
                for rad in display
            ]

        ydata = np.zeros(shape=(npoints, len(display)))
        if yunits in self._get_unit_converter().activity_units:
            for idx in range(0, npoints):
                decayed_contents = self.decay(time_points[idx], xunits).activities(
                    yunits
                )
                ydata[idx] = [decayed_contents[rad] for rad in display]
                ylabel = f"Activity ({yunits})"
        elif yunits in self._get_unit_converter().moles_units:
            for idx in range(0, npoints):
                decayed_contents = self.decay(time_points[idx], xunits).moles(yunits)
                ydata[idx] = [decayed_contents[rad] for rad in display]
                ylabel = f"Number of moles ({yunits})"
        elif yunits in self._get_unit_converter().mass_units:
            for idx in range(0, npoints):
                decayed_contents = self.decay(time_points[idx], xunits).masses(yunits)
                ydata[idx] = [decayed_contents[rad] for rad in display]
                ylabel = f"Mass ({yunits})"
        elif yunits == "num":
            for idx in range(0, npoints):
                decayed_contents = self.decay(time_points[idx], xunits).numbers()
                ydata[idx] = [decayed_contents[rad] for rad in display]
                ylabel = "Number of atoms"
        elif yunits == "activity_frac":
            for idx in range(0, npoints):
                decayed_contents = self.decay(
                    time_points[idx], xunits
                ).activity_fractions()
                ydata[idx] = [decayed_contents[rad] for rad in display]
                ylabel = "Activity fraction"
        elif yunits == "mass_frac":
            for idx in range(0, npoints):
                decayed_contents = self.decay(time_points[idx], xunits).mass_fractions()
                ydata[idx] = [decayed_contents[rad] for rad in display]
                ylabel = "Mass fraction"
        elif yunits == "mol_frac":
            for idx in range(0, npoints):
                decayed_contents = self.decay(time_points[idx], xunits).mole_fractions()
                ydata[idx] = [decayed_contents[rad] for rad in display]
                ylabel = "Mole fraction"
        else:
            raise ValueError(f"{yunits} is not a supported y-axes unit.")

        if yscale == "log" and ymin == 0.0:
            ymin = 0.95 * ydata.min()
        ylimits = [ymin, ymax] if ymax else [ymin, 1.05 * ydata.max()]

        fig, axes = decay_graph(
            time_points=time_points,
            ydata=ydata.T,
            nuclides=display,
            xunits=xunits,
            ylabel=ylabel,
            xscale=xscale,
            yscale=yscale,
            ylimits=ylimits,
            display=set(display),
            fig_in=fig,
            axes_in=axes,
            **kwargs,
        )

        return fig, axes

    def __eq__(self, other: object) -> bool:
        """
        Check whether two instances are equal with ``==`` operator.
        """

        if not isinstance(other, AbstractInventory):
            return NotImplemented
        return self.contents == other.contents and self.decay_data == other.decay_data

    def __ne__(self, other: object) -> bool:
        """
        Check whether two instances are not equal with ``!=`` operator.
        """

        if not isinstance(other, AbstractInventory):
            return NotImplemented
        return not self.__eq__(other)


class Inventory(AbstractInventory):
    # pylint: disable=line-too-long
    """
    ``Inventory`` instances store nuclides and associated quantities (numbers of atoms) in the
    contents dictionary. A decay dataset is associated with each instance.

    Parameters
    ----------
    contents : dict
        Dictionary containing nuclide strings/canonical ids or Nuclide instances as keys and
        quantities (activities, number of atoms, masses or moles) as values.
    units : str, optional
        Units of the contents dictionary values. Specify either 'num' for number of atoms, or an
        activity, mass or moles unit. e.g. 'Bq', 'kBq', 'Ci', 'g', 'kg', 'mol', 'kmol'. Default is
        'Bq'.
    check : bool, optional
        Check for the validity of the contents dictionary (nuclides are in the decay dataset,
        and quantities provided are physical, etc.). Default is True.
    decay_data : DecayData, optional
        Decay dataset (default is the ICRP-107 dataset).

    Attributes
    ----------
    contents : dict
        Dictionary containing nuclide strings as keys and number of atoms of each nuclide as
        values. Nuclides are sorted alphabetically in this dictionary.
    decay_data : DecayData
        Decay dataset.
    decay_matrices : DecayMatricesScipy
        Float/SciPy version of the DecayMatrices associated with the decay dataset.

    Examples
    --------
    >>> rd.Inventory({'Tc-99m': 2.3, 'I-123': 5.8}, 'Bq')
    Inventory activities (Bq): {'I-123': 2.3, 'Tc-99m': 5.8}, decay dataset: icrp107_ame2020_nubase2020
    >>> H3 = rd.Nuclide('H-3')
    >>> rd.Inventory({H3: 3.0}, 'g')
    Inventory activities (Bq): {'H-3': 1067957043281807.0}, decay dataset: icrp107_ame2020_nubase2020
    >>> rd.Inventory({270570000: 7.2, 922380000: 21.1}, 'Ci')
    Inventory activities (Bq): {'Co-57': 266400000000.0, 'U-238': 780700000000.0001}, decay dataset: icrp107_ame2020_nubase2020

    """
    # pylint: enable=line-too-long

    def _get_decay_matrices(self) -> DecayMatricesScipy:
        """Returns the appropriate DecayMatrices instance."""

        return self.decay_data.scipy_data

    @staticmethod
    def _get_quantity_converter() -> Type[QuantityConverter]:
        """
        Returns the appropriate QuantityConverter instance.
        """

        return QuantityConverterFloat

    def _get_unit_converter(self) -> Type[UnitConverter]:
        """
        Returns the appropriate UnitConverter instance.
        """

        return UnitConverterFloat

    def _get_year_conv(self) -> float:
        """Returns the appropriate number of days in a year."""

        return self.decay_data.float_year_conv

    def decay(self, decay_time: float, units: str = "s") -> "Inventory":
        """
        Returns a new inventory calculated from the radioactive decay of the current inventory for
        decay_time.

        Parameters
        ----------
        decay_time : float
            Decay time.
        units : str, optional
            Units of decay_time (default is seconds). Options are 'ns', 'us', 'ms', 's', 'm', 'h',
            'd', 'y', 'ky', 'My', 'Gy', 'Ty', 'Py', and some of the common spelling variations of
            these time units.

        Returns
        -------
        Inventory
            New Inventory after the radioactive decay.

        Examples
        --------
        >>> inv_t0 = rd.Inventory({'H-3': 1.0}, 'Bq')
        >>> inv_t1 = inv_t0.decay(12.32, 'y')
        >>> inv_t1.activities()
        {'H-3': 0.5, 'He-3': 0.0}

        """

        decay_time = self._convert_decay_time(decay_time, units)
        vector_n0, indices, matrix_e = self._setup_decay_calc()

        matrix_e.data[indices] = np.exp(
            -decay_time * self.decay_matrices.decay_consts[indices]
        )

        vector_nt = self._perform_decay_calc(vector_n0, matrix_e)
        new_contents = dict(zip(self.decay_data.nuclides[indices], vector_nt[indices]))

        return self.__class__(new_contents, "num", False, self.decay_data)

    def cumulative_decays(
        self, decay_time: float, units: str = "s"
    ) -> Dict[str, float]:
        """
        Calculates the total number of decays of each nuclide in the inventory between t=0 and
        t=decay_time. Note no results are reported for stable nuclides, as cumulative decays is
        zero.

        Parameters
        ----------
        decay_time : float
            Decay time (calculates total number of decays over this period).
        units : str, optional
            Units of decay_time (default is seconds). Options are 'ns', 'us', 'ms', 's', 'm', 'h',
            'd', 'y', 'ky', 'My', 'Gy', 'Ty', 'Py', and some of the common spelling variations of
            these time units.

        Returns
        -------
        dict
            Dictionary containing nuclide strings as keys and total number of decays of each
            nuclide as values (floats).

        Examples
        --------
        >>> inv_t0 = rd.Inventory({'Sr-90': 10.0}, 'num')
        >>> inv_t0.cumulative_decays(1.0, 'My')
        {'Sr-90': 10.0, 'Y-90': 10.000000000000002}

        """

        decay_time = self._convert_decay_time(decay_time, units)
        vector_n0, indices, matrix_e = self._setup_decay_calc()

        indices = [
            idx for idx in indices if self.decay_matrices.decay_consts[idx] > 0.0
        ]
        for idx in indices:
            matrix_e[idx, idx] = (
                1.0 - np.exp((-decay_time * self.decay_matrices.decay_consts[idx]))
            ) / self.decay_matrices.decay_consts[idx]

        cumulative_decays = self._perform_decay_calc(vector_n0, matrix_e)
        result_dict = {
            self.decay_data.nuclides[idx]: float(
                self.decay_matrices.decay_consts[idx] * cumulative_decays[idx]
            )
            for idx in indices
        }

        return result_dict

    def __repr__(self) -> str:
        return (
            f"Inventory activities (Bq): {self.activities()}, "
            f"decay dataset: {self.decay_data.dataset_name}"
        )


class InventoryHP(AbstractInventory):
    """
    ``InventoryHP`` instances store a dictionary of nuclides and associated numbers of atoms, and a
    ``DecayData`` instance of radioactive decay data. Uses SymPy high precision arithmetic for all
    calculations.

    Parameters
    ----------
    contents : dict
        Dictionary containing nuclide strings/canonical ids or Nuclide
        objects as keys and quantities as values.
    units : str, optional
        Units of the values in the contents dictionary e.g. 'Bq', 'kBq', 'Ci', 'g', 'mol',
        'num'... (default is 'Bq').
    check : bool, optional
        Check for the validity of contents and that the supplied decay dataset contains SymPy data
        (default is True).
    data : DecayData, optional
        Decay dataset (default is the ICRP-107 dataset).
    sympy_contents : dict, optional
        Version of the contents dictionary with SymPy expressions as values. Setting this requires
        that data (the decay dataset) contains SymPy data. ``radioactivedecay`` will create
        sympy_contents automatically from contents if None is supplied.

    Attributes
    ----------
    contents : dict
        Dictionary containing nuclide strings as keys and number of atoms of each nuclide as
        values. Nuclides are sorted alphabetically in this dictionary.
    data : DecayData
        Decay dataset.
    decay_matrices : DecayMatricesSympy
       SymPy DecayMatrices instance associated with the decay dataset.
    sig_fig: int
        Number of significant figures for high precision decay calculations and plots. Deafult is
        320.
    unit_converter : UnitConverterSympy
        SymPy version of a convertor for within different units.

    Examples
    --------
    >>> rd.InventoryHP({'Tc-99m': 2.3, 'I-123': 5.8}, 'Bq')
    InventoryHP activities: {'I-123': 2.3, 'Tc-99m': 5.8}, decay dataset: icrp107
    >>> H3 = rd.Nuclide('H-3')
    >>> rd.InventoryHP({H3: 3.0} 'Bq')
    InventoryHP activities: {'H-3': 3.0}, decay dataset: icrp107
    >>> rd.InventoryHP({'U-238': 21.1, 'Co-57': 7.2}, 'Ci')
    InventoryHP activities: {'Co-57': 7.2, 'U-238': 21.1}, decay dataset: icrp107

    """

    def __init__(
        self,
        contents: Dict[Union[str, int, Nuclide], float],
        units: str = "Bq",
        check: bool = True,
        decay_data: DecayData = DEFAULTDATA,
    ) -> None:
        if check is True:
            try:
                assert decay_data.sympy_data
                assert decay_data.sympy_year_conv
            except ValueError:
                raise ValueError(
                    f"Decay dataset supplied to {self.__class__.__name__} constructor does not "
                    "contain SymPy data."
                ) from None
            contents = {nuc: nsimplify(val) for nuc, val in contents.items()}

        self.sig_fig = 320
        super().__init__(contents, units, check, decay_data)

    def _get_decay_matrices(self) -> DecayMatricesSympy:
        """
        Returns the appropriate DecayMatrices instance.
        """

        return self.decay_data.sympy_data

    @staticmethod
    def _get_quantity_converter() -> Type[QuantityConverter]:
        """
        Returns the appropriate QuantityConverter instance.
        """

        return QuantityConverterSympy

    def _get_unit_converter(self) -> Type[UnitConverter]:
        """
        Returns the appropriate UnitConverter instance.
        """

        return UnitConverterSympy

    def _get_year_conv(self) -> Expr:
        """Returns the appropriate number of days in a year."""

        return self.decay_data.sympy_year_conv

    def numbers(self) -> Dict[str, float]:
        """
        Returns a dictionary containing the number of atoms of each nuclide (as floats) within this
        InventoryHP instance.

        Examples
        --------
        >>> rd.InventoryHP({'Tc-99m': 2.3, 'I-123': 5.8}, 'Bq').numbers()
        {'I-123': 399738.47946141585, 'Tc-99m': 71852.27235544211}

        """

        contents_in_floats = {nuc: float(num) for nuc, num in self.contents.items()}
        return contents_in_floats

    def activities(self, units: str = "Bq") -> Dict[str, float]:
        """
        Returns a dictionary containing the activity of each nuclide (as floats) within this
        InventoryHP instance.

        Parameters
        ----------
        units : str, optional
            Activity units for output, e.g. 'Bq', 'kBq', 'mBq', 'Ci', 'dpm'...
            Deafult is 'Bq'.

        """

        activities = {
            nuc: float(
                self._get_unit_converter().activity_unit_conv(
                    self._get_quantity_converter().number_to_activity(
                        num, self._get_decay_const(nuc)
                    ),
                    "Bq",
                    units,
                )
            )
            for nuc, num in self.contents.items()
        }

        return activities

    def masses(self, units: str = "g") -> Dict[str, float]:
        """
        Returns a dictionary containing the mass of each nuclide (as floats) within this InventoryHP
        instance.

        Parameters
        ----------
        units : str, optional
            Mass units for output, e.g. 'Bq', 'g', 'kg', 'mg', 'ton'...
            Deafult is 'g'.

        """

        masses = {
            nuc: float(
                self._get_unit_converter().mass_unit_conv(
                    self._get_quantity_converter().number_to_mass(
                        num, self._get_atomic_mass(nuc)
                    ),
                    "g",
                    units,
                )
            )
            for nuc, num in self.contents.items()
        }

        return masses

    def moles(self, units: str = "mol") -> Dict[str, float]:
        """
        Returns a dictionary containing the number of atoms of each nuclide (as floats) within this
        InventoryHP instance.

        Parameters
        ----------
        units : str, optional
            Moles units, e.g. 'mmol', 'mol', 'kmol'...
            Deafult is 'mol'.

        """

        moles = {
            nuc: float(
                self._get_unit_converter().moles_unit_conv(
                    self._get_quantity_converter().number_to_moles(num), "mol", units
                )
            )
            for nuc, num in self.contents.items()
        }

        return moles

    def decay(self, decay_time: float, units: str = "s") -> "InventoryHP":
        """
        Decay calculation with high numerical precision. This uses SymPy arbitrary-precision
        arithmetic functions for the decay calculation. The results can be more accurate than
        a normal double precision float calculation with the ``Inventory`` class when the decay
        chains contain radionuclides with very similar half-lives or half-lives that differ by many
        orders of magnitude.

        Parameters
        ----------
        decay_time : float
            Decay time.
        units : str, optional
            Units of decay_time (default is seconds). Options are 'ns', 'us', 'ms', 's', 'm', 'h',
            'd', 'y', 'ky', 'My', 'Gy', 'Ty', 'Py', and some of the common spelling variations of
            these time units.

        Returns
        -------
        InventoryHP
            New high precision inventory after the radioactive decay.

        Raises
        ------
        ValueError
            If self.sig_fig is set to be lower than 1.

        Examples
        --------
        >>> inv_t0 = rd.InventoryHP({'Fm-257': 1.0})
        >>> inv_t1 = inv_t0.decay(10.0, 'd')
        >>> inv_t1.activities()
        {'Ac-225': 0.0,
        'Am-241': 9.985270042416324e-24,
        'Am-245': 5.4061684195880344e-09,
        ...
        'Fm-257': 0.9333548028364793,
        ...
        }

        """

        if self.sig_fig < 1:
            raise ValueError(
                "InventoryHP.sig_fig attribute needs to be int greater than 0."
            )

        decay_time = nsimplify(decay_time)
        decay_time = self._convert_decay_time(decay_time, units)
        vector_n0, indices, matrix_e = self._setup_decay_calc()

        for idx in indices:
            matrix_e[idx, idx] = exp(
                (-decay_time * self.decay_matrices.decay_consts[idx]).evalf(
                    self.sig_fig
                )
            )

        vector_nt = self._perform_decay_calc(vector_n0, matrix_e)
        new_contents = {
            self.decay_data.nuclides[idx]: vector_nt[idx] for idx in indices
        }

        return self.__class__(new_contents, "num", False, self.decay_data)

    def cumulative_decays(
        self, decay_time: float, units: str = "s"
    ) -> Dict[str, float]:
        """
        Calculates the total number of decays of each nuclide in the inventory between t=0 and
        t=decay_time. Uses SymPy high precision calculations. Note no results are reported for
        stable nuclides, as cumulative decays is zero.

        Parameters
        ----------
        decay_time : float
            Decay time (calculates total number of decays over this period).
        units : str, optional
            Units of decay_time (default is seconds). Options are 'ns', 'us', 'ms', 's', 'm', 'h',
            'd', 'y', 'ky', 'My', 'Gy', 'Ty', 'Py', and some of the common spelling variations of
            these time units.

        Returns
        -------
        dict
            Dictionary containing nuclide strings as keys and total number of decays of each
            nuclide as values (floats).

        Raises
        ------
        ValueError
            If self.sig_fig is set to be lower than 1.

        Examples
        --------
        >>> inv_t0 = rd.Inventory({'Sr-90': 10.0}, 'num')
        >>> inv_t0.cumulative_decays(1.0, 'My')
        {'Sr-90': 10.0, 'Y-90': 10.0}

        """

        if self.sig_fig < 1:
            raise ValueError(
                "InventoryHP.sig_fig attribute needs to be int greater than 0."
            )

        decay_time = nsimplify(decay_time)
        decay_time = self._convert_decay_time(decay_time, units)
        vector_n0, indices, matrix_e = self._setup_decay_calc()

        indices = [
            idx for idx in indices if self.decay_matrices.decay_consts[idx] > Integer(0)
        ]
        for idx in indices:
            matrix_e[idx, idx] = (
                Integer(1)
                - exp(
                    (-decay_time * self.decay_matrices.decay_consts[idx]).evalf(
                        self.sig_fig
                    )
                )
            ) / self.decay_matrices.decay_consts[idx]

        cumulative_decays = self._perform_decay_calc(vector_n0, matrix_e)
        result_dict = {
            self.decay_data.nuclides[idx]: float(
                self.decay_matrices.decay_consts[idx] * cumulative_decays[idx]
            )
            for idx in indices
        }

        return result_dict

    # Redefines one default parameter of the plot function (npoints) for performance reasons.
    # Note rewriting all the parameters like this violates the Don't Repeat Yourself principle, but
    # currently no better solution while still maintaining auto-complete functionality for
    # the method signature with some editors.
    def plot(  # type: ignore
        self,
        xmax: float,
        xunits: str = "s",
        xmin: float = 0.0,
        xscale: str = "linear",
        yscale: str = "linear",
        ymin: float = 0.0,
        ymax: Optional[float] = None,
        yunits: str = "Bq",
        display: Union[str, List[str]] = "all",
        order: str = "dataset",
        npoints: int = 51,
        fig: Optional[matplotlib.figure.Figure] = None,
        axes: Optional[matplotlib.axes.Axes] = None,
        **kwargs,
    ) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
        """
        Plots a decay graph using high precision decay calculations. Only difference from normal
        precision decay plot (``Inventory.plot()``) is default npoints=51.

        Parameters
        ----------
        xmax : float
            Maximum decay time on x-axis.
        xunits : str, optional
            Units for decay times (default is 's', i.e. seconds). Options are 'ps', 'ns', 'us',
            'ms', 's', 'm', 'h', 'd', 'y', 'ky', 'My', 'Gy', 'Ty', 'Py', and some of the common
            spelling variations of these time units.
        xmin : float, optional
            Minimum decay time on x-axis (default is 0.0 for linear x-axis, 0.1 for log x-axis).
        xscale : str, optional
            The time axis scale type to apply ('linear' or 'log', default is 'linear').
        yscale : str, optional
            The y-axis scale type to apply ('linear' or 'log', default is 'linear').
        ymin : float, optional
            Minimum value for the y-axis (default is 0.0 for ``yscale='linear'``, or 0.95x the
            minimum quantity that occurs over the decay period for ``yscale='log'``).
        ymax : None or float, optional
            Maximum value for the y-axis. Default is None, which sets the limit to 1.05x the
            maximum quantity that occurs over the decay period.
        yunits : str, optional
            Units to display on the y-axis e.g. 'Bq', 'kBq', 'Ci', 'g', 'mol', 'num',
            'activity_frac', 'mass_frac', 'mol_frac'. Default is 'Bq'.
        display : str or list, optional
            Only display the nuclides within this list on the graph. Use this parameter when
            you want to choose specific nuclide decay curves shown on the graph, either by
            supplying a string (to show one nuclide) or a list of strings (to show multiple).
            Default is 'all', which displays all nuclides present upon decay of the inventory.
        order : str, optional
            Order to display the nuclide decay curves on the graph if you do not specify the
            order via the display parameter. Default order is by 'dataset', which follows the order
            of the nuclides in the decay dataset (highest to lowest nuclides in the decay
            chains). Use 'alphabetical' if you want the nuclides to be ordered alphabetically.
        npoints : None or int, optional
            Number of time points used to plot graph. Default is 51.
        fig : None or matplotlib.figure.Figure, optional
            matplotlib figure object to use, or None makes ``radioactivedecay`` create one (default
            is None).
        axes : None or matplotlib.axes.Axes, optional
            matplotlib axes object to use, or None makes ``radioactivedecay`` create one (default
            is None).
        **kwargs, optional
            All additional keyword arguments to supply to matplotlib plot() function.

        Returns
        -------
        fig : matplotlib.figure.Figure
            matplotlib figure object used to plot decay chain.
        axes : matplotlib.axes.Axes
            matplotlib axes object used to plot decay chain.

        Raises
        ------
        ValueError
            If the order parameter is invalid.

        """

        return super().plot(
            xmax,
            xunits,
            xmin,
            xscale,
            yscale,
            ymin,
            ymax,
            yunits,
            display,
            order,
            npoints,
            fig,
            axes,
            **kwargs,
        )

    def __repr__(self) -> str:
        return (
            f"InventoryHP activities (Bq): {self.activities()}, "
            f"decay dataset: {self.decay_data.dataset_name}"
        )



"""
Unit tests for utils.py functions.
"""

import unittest

import numpy as np
from sympy import Integer, log

from radioactivedecay.utils import (
    NuclideStrError,
    Z_to_elem,
    add_dictionaries,
    build_id,
    build_nuclide_string,
    elem_to_Z,
    get_metastable_chars,
    parse_id,
    parse_nuclide,
    parse_nuclide_str,
    sort_dictionary_alphabetically,
    sort_list_according_to_dataset,
)


class TestFunctions(unittest.TestCase):
    """
    Unit tests for the utils.py functions.
    """

    def test_get_metastable_chars(self) -> None:
        """
        Test fetching of list of metastable state characters.
        """

        self.assertEqual(get_metastable_chars(), ["m", "n", "p", "q", "r", "x"])

    def test_Z_to_elem(self) -> None:
        """
        Test the conversion of atomic number to element symbol.
        """

        self.assertEqual(Z_to_elem(1), "H")
        self.assertEqual(Z_to_elem(20), "Ca")
        self.assertEqual(Z_to_elem(26), "Fe")

    def test_elem_to_Z(self) -> None:
        """
        Test the conversion of element symbol to atomic number.
        """

        self.assertEqual(elem_to_Z("H"), 1)
        self.assertEqual(elem_to_Z("Ca"), 20)
        self.assertEqual(elem_to_Z("Fe"), 26)

    def test_build_id(self) -> None:
        """
        Test the canonical id builder.
        """

        self.assertEqual(build_id(26, 56), 260560000)
        self.assertEqual(build_id(53, 118), 531180000)
        self.assertEqual(build_id(53, 118, "m"), 531180001)
        self.assertEqual(build_id(65, 156, "n"), 651560002)
        self.assertEqual(build_id(49, 129, "p"), 491290003)
        self.assertEqual(build_id(71, 177, "q"), 711770004)
        self.assertEqual(build_id(71, 177, "r"), 711770005)
        self.assertEqual(build_id(71, 174, "x"), 711740006)

        with self.assertRaises(ValueError):
            build_id(65, 156, "z")

    def test_built_nuclide_string(self) -> None:
        """
        Test the nuclide string builder.
        """

        self.assertEqual(build_nuclide_string(26, 56), "Fe-56")
        self.assertEqual(build_nuclide_string(53, 118), "I-118")
        self.assertEqual(build_nuclide_string(53, 118, "m"), "I-118m")
        self.assertEqual(build_nuclide_string(65, 156, "n"), "Tb-156n")
        self.assertEqual(build_nuclide_string(49, 129, "p"), "In-129p")
        self.assertEqual(build_nuclide_string(71, 177, "q"), "Lu-177q")
        self.assertEqual(build_nuclide_string(71, 177, "r"), "Lu-177r")
        self.assertEqual(build_nuclide_string(71, 174, "x"), "Lu-174x")

        with self.assertRaises(ValueError):
            build_nuclide_string(999, 1000, "z")

    def test_parse_nuclide_str(self) -> None:
        """
        Test the parsing of nuclide strings.
        """

        self.assertEqual(parse_nuclide_str("Ca-40"), "Ca-40")
        self.assertEqual(parse_nuclide_str("Ca40"), "Ca-40")
        self.assertEqual(parse_nuclide_str("40Ca"), "Ca-40")

        # Whitespace removal (Issue #65)
        self.assertEqual(parse_nuclide_str(" Ca -40 "), "Ca-40")
        self.assertEqual(parse_nuclide_str("C\ta\n-40"), "Ca-40")

        # Robust to capitalization mistakes (Issue #65)
        self.assertEqual(parse_nuclide_str("y-91"), "Y-91")
        self.assertEqual(parse_nuclide_str("y91"), "Y-91")
        self.assertEqual(parse_nuclide_str("91y"), "Y-91")
        self.assertEqual(parse_nuclide_str("y-91M"), "Y-91m")
        self.assertEqual(parse_nuclide_str("y91M"), "Y-91m")
        # Following test will fail as no capitalization of Y
        # self.assertEqual(parse_nuclide_str("91my"), "Y-91m")
        self.assertEqual(parse_nuclide_str("ca-40"), "Ca-40")
        self.assertEqual(parse_nuclide_str("CA-40"), "Ca-40")
        self.assertEqual(parse_nuclide_str("Tc-99M"), "Tc-99m")
        self.assertEqual(parse_nuclide_str("iR192N"), "Ir-192n")
        self.assertEqual(parse_nuclide_str("192NiR"), "Ir-192n")
        self.assertEqual(parse_nuclide_str("iN129P"), "In-129p")
        self.assertEqual(parse_nuclide_str("177qLu"), "Lu-177q")
        self.assertEqual(parse_nuclide_str("LU177R"), "Lu-177r")
        self.assertEqual(parse_nuclide_str("lu-174x"), "Lu-174x")

        self.assertEqual(parse_nuclide_str("ni56"), "Ni-56")
        self.assertEqual(parse_nuclide_str("ni-56"), "Ni-56")
        self.assertEqual(parse_nuclide_str("56Ni"), "Ni-56")
        self.assertEqual(parse_nuclide_str("56ni"), "Ni-56")
        # Following test will fail as logic assumes this is I-56n
        # self.assertEqual(parse_nuclide_str("56nI"), "Ni-56")
        self.assertEqual(parse_nuclide_str("ni69M"), "Ni-69m")
        self.assertEqual(parse_nuclide_str("ni-69n"), "Ni-69n")
        self.assertEqual(parse_nuclide_str("69nni"), "Ni-69n")

        self.assertEqual(parse_nuclide_str("130nI"), "I-130n")
        # Following tests will fail as logic assumes Ni-130
        # self.assertEqual(parse_nuclide_str("130NI"), "I-130n")
        # self.assertEqual(parse_nuclide_str("130Ni"), "I-130n")
        # self.assertEqual(parse_nuclide_str("130ni"), "I-130n")

        with self.assertRaises(NuclideStrError):
            parse_nuclide_str("H3.")  # not alpha-numeric
        with self.assertRaises(NuclideStrError):
            parse_nuclide_str("H-3-")  # too many hyphens
        with self.assertRaises(NuclideStrError):
            parse_nuclide_str("H-301")  # mass number too large
        with self.assertRaises(NuclideStrError):
            parse_nuclide_str("H")  # no mass number
        with self.assertRaises(NuclideStrError):
            parse_nuclide_str("Tc-99m3")  # more than one number
        with self.assertRaises(NuclideStrError):
            parse_nuclide_str("F26m0")  # more than one number
        with self.assertRaises(NuclideStrError):
            parse_nuclide_str("A3")  # invalid element
        with self.assertRaises(NuclideStrError):
            parse_nuclide_str("Tc-99mm")  # metastable char too long
        with self.assertRaises(NuclideStrError):
            parse_nuclide_str("Tc-99o")  # metastable char invalid

    def test_parse_id(self) -> None:
        """
        Test the canonical id to nuclide string converter.
        """

        self.assertEqual(parse_id(260560000), "Fe-56")
        self.assertEqual(parse_id(531180000), "I-118")
        self.assertEqual(parse_id(531180001), "I-118m")
        self.assertEqual(parse_id(651560002), "Tb-156n")
        self.assertEqual(parse_id(491290003), "In-129p")
        self.assertEqual(parse_id(711770004), "Lu-177q")
        self.assertEqual(parse_id(711770005), "Lu-177r")
        self.assertEqual(parse_id(711740006), "Lu-174x")

    def test_parse_nuclide(self) -> None:
        """
        Test the parsing of nuclide strings.
        """

        nuclides = np.array(
            [
                "H-3",
                "Be-7",
                "C-10",
                "Ne-19",
                "I-118",
                "Pd-100",
                "Cl-34m",
                "I-118m",
                "Tb-156m",
                "Tb-156n",
                "In-129p",
                "Lu-177q",
                "Lu-177r",
                "Lu-174x",
            ]
        )
        dataset_name = "test"

        # Re-formatting of acceptable strings e.g. 100Pd -> Pd-100
        self.assertEqual(parse_nuclide("H-3", nuclides, dataset_name), "H-3")
        self.assertEqual(parse_nuclide("H3", nuclides, dataset_name), "H-3")
        self.assertEqual(parse_nuclide("3H", nuclides, dataset_name), "H-3")
        self.assertEqual(parse_nuclide(10030000, nuclides, dataset_name), "H-3")
        self.assertEqual(parse_nuclide("Be-7", nuclides, dataset_name), "Be-7")
        self.assertEqual(parse_nuclide("Be7", nuclides, dataset_name), "Be-7")
        self.assertEqual(parse_nuclide("7Be", nuclides, dataset_name), "Be-7")
        self.assertEqual(parse_nuclide(40070000, nuclides, dataset_name), "Be-7")
        self.assertEqual(parse_nuclide("C-10", nuclides, dataset_name), "C-10")
        self.assertEqual(parse_nuclide("C10", nuclides, dataset_name), "C-10")
        self.assertEqual(parse_nuclide("10C", nuclides, dataset_name), "C-10")
        self.assertEqual(parse_nuclide(60100000, nuclides, dataset_name), "C-10")
        self.assertEqual(parse_nuclide("Ne-19", nuclides, dataset_name), "Ne-19")
        self.assertEqual(parse_nuclide("Ne19", nuclides, dataset_name), "Ne-19")
        self.assertEqual(parse_nuclide("19Ne", nuclides, dataset_name), "Ne-19")
        self.assertEqual(parse_nuclide(100190000, nuclides, dataset_name), "Ne-19")
        self.assertEqual(parse_nuclide("I-118", nuclides, dataset_name), "I-118")
        self.assertEqual(parse_nuclide("I118", nuclides, dataset_name), "I-118")
        self.assertEqual(parse_nuclide("118I", nuclides, dataset_name), "I-118")
        self.assertEqual(parse_nuclide(531180000, nuclides, dataset_name), "I-118")
        self.assertEqual(parse_nuclide("Pd-100", nuclides, dataset_name), "Pd-100")
        self.assertEqual(parse_nuclide("Pd100", nuclides, dataset_name), "Pd-100")
        self.assertEqual(parse_nuclide("100Pd", nuclides, dataset_name), "Pd-100")
        self.assertEqual(parse_nuclide(461000000, nuclides, dataset_name), "Pd-100")
        self.assertEqual(parse_nuclide("Cl-34m", nuclides, dataset_name), "Cl-34m")
        self.assertEqual(parse_nuclide("Cl34m", nuclides, dataset_name), "Cl-34m")
        self.assertEqual(parse_nuclide("34mCl", nuclides, dataset_name), "Cl-34m")
        self.assertEqual(parse_nuclide(170340001, nuclides, dataset_name), "Cl-34m")
        self.assertEqual(parse_nuclide("I-118m", nuclides, dataset_name), "I-118m")
        self.assertEqual(parse_nuclide("I118m", nuclides, dataset_name), "I-118m")
        self.assertEqual(parse_nuclide("118mI", nuclides, dataset_name), "I-118m")
        self.assertEqual(parse_nuclide(531180001, nuclides, dataset_name), "I-118m")
        self.assertEqual(parse_nuclide("Tb-156m", nuclides, dataset_name), "Tb-156m")
        self.assertEqual(parse_nuclide("Tb156m", nuclides, dataset_name), "Tb-156m")
        self.assertEqual(parse_nuclide("156mTb", nuclides, dataset_name), "Tb-156m")
        self.assertEqual(parse_nuclide(651560001, nuclides, dataset_name), "Tb-156m")
        self.assertEqual(parse_nuclide("Tb-156n", nuclides, dataset_name), "Tb-156n")
        self.assertEqual(parse_nuclide("Tb156n", nuclides, dataset_name), "Tb-156n")
        self.assertEqual(parse_nuclide("156nTb", nuclides, dataset_name), "Tb-156n")
        self.assertEqual(parse_nuclide(651560002, nuclides, dataset_name), "Tb-156n")
        self.assertEqual(parse_nuclide("In-129p", nuclides, dataset_name), "In-129p")
        self.assertEqual(parse_nuclide("In129p", nuclides, dataset_name), "In-129p")
        self.assertEqual(parse_nuclide("129pIn", nuclides, dataset_name), "In-129p")
        self.assertEqual(parse_nuclide(491290003, nuclides, dataset_name), "In-129p")
        self.assertEqual(parse_nuclide("Lu-177q", nuclides, dataset_name), "Lu-177q")
        self.assertEqual(parse_nuclide("Lu177q", nuclides, dataset_name), "Lu-177q")
        self.assertEqual(parse_nuclide("177qLu", nuclides, dataset_name), "Lu-177q")
        self.assertEqual(parse_nuclide(711770004, nuclides, dataset_name), "Lu-177q")
        self.assertEqual(parse_nuclide("Lu-177r", nuclides, dataset_name), "Lu-177r")
        self.assertEqual(parse_nuclide("Lu-177r", nuclides, dataset_name), "Lu-177r")
        self.assertEqual(parse_nuclide("177rLu", nuclides, dataset_name), "Lu-177r")
        self.assertEqual(parse_nuclide(711770005, nuclides, dataset_name), "Lu-177r")
        self.assertEqual(parse_nuclide("Lu-174x", nuclides, dataset_name), "Lu-174x")
        self.assertEqual(parse_nuclide("Lu-174x", nuclides, dataset_name), "Lu-174x")
        self.assertEqual(parse_nuclide("174xLu", nuclides, dataset_name), "Lu-174x")
        self.assertEqual(parse_nuclide(711740006, nuclides, dataset_name), "Lu-174x")

        # Catch erroneous strings
        with self.assertRaises(TypeError):
            parse_nuclide(1.2, nuclides, dataset_name)
        with self.assertRaises(ValueError):
            parse_nuclide("H", nuclides, dataset_name)
        with self.assertRaises(ValueError):
            parse_nuclide("A1", nuclides, dataset_name)
        with self.assertRaises(ValueError):
            parse_nuclide("1A", nuclides, dataset_name)
        with self.assertRaises(ValueError):
            parse_nuclide("H-4", nuclides, dataset_name)
        with self.assertRaises(ValueError):
            parse_nuclide("H4", nuclides, dataset_name)
        with self.assertRaises(ValueError):
            parse_nuclide("4H", nuclides, dataset_name)
        with self.assertRaises(ValueError):
            parse_nuclide("Pb-198m", nuclides, dataset_name)
        with self.assertRaises(ValueError):
            parse_nuclide("Pbo-198m", nuclides, dataset_name)

    def test_add_dictionaries(self) -> None:
        """
        Test function which adds two inventory dictionaries together.
        """

        dict1 = {"Pm-141": 1.0, "Rb-78": 2.0}
        dict2 = {"Pm-141": 3.0, "Rb-90": 4.0}
        self.assertEqual(
            add_dictionaries(dict1, dict2),
            {"Pm-141": 4.0, "Rb-78": 2.0, "Rb-90": 4.0},
        )

        dict1 = {"Pm-141": Integer(2) * log(3), "Rb-78": Integer(4) / log(5)}
        dict2 = {"Pm-141": log(3) / Integer(7), "Rb-90": Integer(9)}
        self.assertEqual(
            add_dictionaries(dict1, dict2),
            {
                "Pm-141": Integer(15) * log(3) / Integer(7),
                "Rb-78": Integer(4) / log(5),
                "Rb-90": Integer(9),
            },
        )

    def test_sort_dictionary_alphabetically(self) -> None:
        """
        Test the sorting of a dictionary by its keys alphabetically.
        """

        inv_dict = {"U-235": 1.2, "Tc-99m": 2.3, "Tc-99": 5.8}
        self.assertEqual(
            sort_dictionary_alphabetically(inv_dict),
            {"Tc-99": 5.8, "Tc-99m": 2.3, "U-235": 1.2},
        )

        inv_dict = {"U-235": Integer(1), "Tc-99m": Integer(2), "Tc-99": Integer(3)}
        self.assertEqual(
            sort_dictionary_alphabetically(inv_dict),
            {"Tc-99": Integer(3), "Tc-99m": Integer(2), "U-235": Integer(1)},
        )

    def test_sort_list_according_to_dataset(self) -> None:
        """
        Test the sorting of list of nuclides according to their position in the decay dataset.
        """

        nuclide_list = ["Tc-99", "Tc-99m"]
        nuclide_dict = {"Tc-99m": 0, "Tc-99": 1}
        self.assertEqual(
            sort_list_according_to_dataset(nuclide_list, nuclide_dict),
            ["Tc-99m", "Tc-99"],
        )


class TestNuclideStrError(unittest.TestCase):
    """
    Unit tests for the NuclideStrError class.
    """

    def test_instantiation(self) -> None:
        """
        Test instantiation of NuclideStrError exceptions.
        """

        err = NuclideStrError("A4", "Dummy message.")
        self.assertEqual(err.nuclide, "A4")
        self.assertEqual(err.additional_message, "Dummy message.")

    def test___str__(self) -> None:
        """
        Test string representation f NuclideStrError exceptions.
        """

        err = NuclideStrError("A4", "Dummy message.")
        self.assertEqual(str(err), "A4 is not a valid nuclide string. Dummy message.")


if __name__ == "__main__":
    unittest.main()



"""
The nuclide module defines the ``Nuclide`` class. Each ``Nuclide`` instance
represents one nuclide, and can be contructed from a nuclide string or
`zzzaaassss` canonical id. The class properties provide an easy way of
extracting atomic number and mass information, as well as a clean name string.
Furthermore, additional methods provide an access point for mass data of
nuclides, and the decay data of radionuclides, if present in a specified
dataset. The default decay dataset used if none is supplied to the constructor
is rd.DEFAULTDATA.

The docstring code examples assume that ``radioactivedecay`` has been imported
as ``rd``:

.. highlight:: python
.. code-block:: python

    >>> import radioactivedecay as rd

"""

from collections import deque
from typing import Any, Dict, List, Optional, Tuple, Union

import matplotlib
import networkx as nx
import numpy as np

from radioactivedecay.decaydata import DEFAULTDATA, DecayData
from radioactivedecay.plots import (
    _check_fig_axes,
    _parse_decay_mode_label,
    _parse_nuclide_label,
)
from radioactivedecay.utils import build_id, elem_to_Z, parse_nuclide


class Nuclide:
    """
    ``Nuclide`` instances serve as name and atomic number/mass parsing
    objects for any nuclide or element.

    Parameters
    ----------
    nuclide : str or int
        Specify the nuclide with either a name string (e.g. 'H-3', 'H3' or '3H') or a canonical id
        (zzzaaassss format).
    decay_data : DecayData, optional
        Decay dataset (default is the ICRP-107 dataset).

    Attributes
    ----------
    nuclide: str
        Nuclide name string.
    decay_data : DecayData
        Decay dataset.

    Examples
    --------
    >>> rd.Nuclide('K-40')
    Nuclide: K-40, decay dataset: icrp107_ame2020_nubase2020
    >>> rd.Nuclide('K40')
    Nuclide: K-40, decay dataset: icrp107_ame2020_nubase2020
    >>> rd.Nuclide(190400000)
    Nuclide: K-40, decay dataset: icrp107_ame2020_nubase2020
    >>> rd.Nuclide(280560001)
    Nuclide: Ni-56m, decay dataset: icrp107_ame2020_nubase2020

    """

    def __init__(
        self, nuclide: Union[str, int], decay_data: DecayData = DEFAULTDATA
    ) -> None:
        self.decay_data = decay_data
        self.nuclide = parse_nuclide(
            nuclide, self.decay_data.nuclides, self.decay_data.dataset_name
        )

    @property
    def Z(self) -> int:
        """
        Returns the atomic number of the nuclide.

        Returns
        -------
        int
            Atomic number of the nuclide.

        Examples
        --------
        >>> H3 = rd.Nuclide('H-3')
        >>> H3.Z
        1

        """

        return elem_to_Z(self.nuclide.split("-")[0])

    @property
    def A(self) -> int:
        """
        Returns the mass number of the nuclide.

        Returns
        -------
        int
            Mass number of the nuclide.

        Examples
        --------
        >>> H3 = rd.Nuclide('H-3')
        >>> H3.A
        3

        """

        return int(self.nuclide.split("-")[1].strip("mn"))

    @property
    def state(self) -> str:
        """
        Returns the metastable state character, i.e. '' for ground state, 'm' for first metastable
        state, 'n' for second metastable state, etc..

        Returns
        -------
        str
            Metastable state character.

        Examples
        --------
        >>> H3 = rd.Nuclide('H-3')
        >>> H3.state
        ''
        >>> Ba137m = rd.Nuclide('Ba-137m')
        >>> Ba137m.state
        'm'

        """

        return self.nuclide.split("-")[1].strip("0123456789")

    @property
    def id(self) -> int:
        """
        Returns the canonical nuclide id, in zzzaaassss form. Ground state is 0000, first excited
        state ("m") is 0001, second ("n") is 0002, etc.

        Returns
        -------
        int
            Canonical id of the nuclide.

        Examples
        --------
        >>> H3 = rd.Nuclide('H-3')
        >>> H3.id
        10030000
        >>> Ba137m = rd.Nuclide('Ba-137m')
        >>> Ba137m.id
        561370001

        """

        return build_id(self.Z, self.A, self.state)

    @property
    def atomic_mass(self) -> float:
        """
        Returns the atomic mass of the nuclide, in g/mol.

        Returns
        -------
        float
            Atomic mass of the nuclide in g/mol.

        Examples
        --------
        >>> H3 = rd.Nuclide('H-3')
        >>> H3.atomic_mass
        3.01604928132
        >>> Ba137m = rd.Nuclide('Ba-137m')
        >>> Ba137m.atomic_mass
        136.9065375271172

        """

        return self.decay_data.scipy_data.atomic_masses[
            self.decay_data.nuclide_dict[self.nuclide]
        ]

    def half_life(self, units: str = "s") -> Union[float, str]:
        """
        Returns the half-life of a nuclide as a float in your chosen
        units, or as a human-readable string with appropriate units.

        Parameters
        ----------
        units : str, optional
            Units for half-life. Options are 'ps', 'ns', 'μs', 'us',
            'ms', 's', 'm', 'h', 'd', 'y', 'ky', 'My', 'By', 'Gy',
            'Ty', 'Py', and common spelling variations. Default is 's',
            i.e. seconds. Use 'readable' to get a string of the
            half-life in human-readable units.

        Returns
        -------
        float or str
            Half-life of the nuclide.

        Examples
        --------
        >>> K40 = rd.Nuclide('K-40')
        >>> K40.half_life('y')
        1251000000.0
        >>> K40.half_life('readable')
        '1.251 By'
        >>> Fe56 = rd.Nuclide('Fe-56')
        >>> Fe56.half_life('readable')
        'stable'

        """

        return self.decay_data.half_life(self.nuclide, units)

    def progeny(self) -> List[str]:
        """
        Returns list of the direct progeny of the nuclide.

        Returns
        -------
        list
            List of the direct progeny of the nuclide, ordered by decreasing branching fraction.

        Examples
        --------
        >>> K40 = rd.Nuclide('K-40')
        >>> K40.progeny()
        ['Ca-40', 'Ar-40']

        """

        return self.decay_data.progeny[self.decay_data.nuclide_dict[self.nuclide]]

    def branching_fractions(self) -> List[float]:
        """
        Returns the branching fractions to the direct progeny of a
        radionuclide.

        Returns
        -------
        list
            List of branching fractions.

        Examples
        --------
        >>> K40 = rd.Nuclide('K-40')
        >>> K40.branching_fractions()
        [0.8914, 0.1086]

        """

        return self.decay_data.bfs[self.decay_data.nuclide_dict[self.nuclide]]

    def decay_modes(self) -> List[str]:
        """
        Returns the decay modes for a radionuclide, as defined in the
        decay dataset. Note: the decay mode strings returned are not
        lists of all the different radiation types emitted during the
        parent to progeny decay processes. They are the labels defined
        in the decay dataset to classify the parent to progeny decay
        type (e.g. '\u03b1', '\u03b2-' or 'IT').

        Returns
        -------
        list
            List of decay modes.

        Examples
        --------
        >>> K40 = rd.Nuclide('K-40')
        >>> K40.decay_modes()
        ['\u03b2-', '\u03b2+ & EC']

        """

        return self.decay_data.modes[self.decay_data.nuclide_dict[self.nuclide]]

    def plot(
        self,
        label_pos: float = 0.5,
        fig: Optional[matplotlib.figure.Figure] = None,
        axes: Optional[matplotlib.axes.Axes] = None,
        kwargs_draw: Optional[Dict[str, Any]] = None,
        kwargs_edge_labels: Optional[Dict[str, Any]] = None,
    ) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
        """
        Plots a diagram of the decay chain of a radionuclide. Then
        creates a NetworkX DiGraph and plot of it using NetworkX's
        Matplotlib-based plotting functionality.

        Some of the NetworkX default plotting parameters are changed to
        produce nice decay chain diagrams. However, users retain
        control over these parameters via kwargs_draw and
        kwargs_edge_labels. For more information on the various
        NetworkX plotting parameters, refer to its `documentation
        <https://networkx.org/documentation/stable/reference/drawing.html>`_.

        Parameters
        ----------
        label_pos : float, optional
            Position of labels along edges. Default is 0.5. If you find
            that edge labels are overlapping in the decay chain
            diagram, try increasing this parameter to e.g. 0.66.
        fig : None or matplotlib.figure.Figure, optional
            matplotlib figure object to use, or None makes
            ``radioactivedecay`` create one (default is None).
        axes : None or matplotlib.axes.Axes, optional
            matplotlib axes object to use, or None makes
            ``radioactivedecay`` create one (default is None).
        **kwargs_draw, optional
            Keyword arguments for networkx.draw().
        **kwargs_edge_labels, optional
            Keyword arguments for networkx.draw_networkx_edge_labels().

        Returns
        -------
        fig : matplotlib.figure.Figure
            matplotlib figure object used to plot the decay chain.
        ax : matplotlib.axes.Axes
            matplotlib axes object used to plot the decay chain.

        """

        digraph, max_generation, max_xpos = _build_decay_digraph(self, nx.DiGraph())

        positions = nx.get_node_attributes(digraph, "pos")
        node_labels = nx.get_node_attributes(digraph, "label")
        edge_labels = nx.get_edge_attributes(digraph, "label")

        fig, axes = _check_fig_axes(
            fig, axes, figsize=(3 * max_xpos + 1.5, 3 * max_generation + 1.5)
        )

        if kwargs_draw is None:
            kwargs_draw = {}
        if "node_size" not in kwargs_draw:
            kwargs_draw["node_size"] = 6000
        if "node_color" not in kwargs_draw:
            kwargs_draw["node_color"] = "#FFFFFF"
        if "edgecolors" not in kwargs_draw:
            kwargs_draw["edgecolors"] = "#000000"

        nx.draw(
            G=digraph,
            pos=positions,
            ax=axes,
            labels=node_labels,
            **kwargs_draw,
        )

        if kwargs_edge_labels is None:
            kwargs_edge_labels = {}
        if "font_size" not in kwargs_edge_labels:
            kwargs_edge_labels["font_size"] = 12
        if "bbox" not in kwargs_edge_labels:
            kwargs_edge_labels["bbox"] = dict(
                boxstyle=None, ec=(1.0, 1.0, 1.0), fc=(1.0, 1.0, 1.0)
            )
        if "rotate" not in kwargs_edge_labels:
            kwargs_edge_labels["rotate"] = False

        nx.draw_networkx_edge_labels(
            G=digraph,
            pos=positions,
            edge_labels=edge_labels,
            label_pos=label_pos,
            ax=axes,
            **kwargs_edge_labels,
        )

        axes.set_xlim(-0.3, max_xpos + 0.3)
        axes.set_ylim(-max_generation - 0.3, 0.3)

        return fig, axes

    def __repr__(self) -> str:
        rep = f"Nuclide: {self.nuclide}, decay dataset: {self.decay_data.dataset_name}"

        return rep

    def __eq__(self, other: object) -> bool:
        """
        Check whether two ``Nuclide`` instances are equal with ``==`` operator.
        """

        if not isinstance(other, Nuclide):
            return NotImplemented
        return self.nuclide == other.nuclide and self.decay_data == other.decay_data

    def __ne__(self, other: object) -> bool:
        """
        Check whether two ``Nuclide`` instances are not equal with ``!=`` operator.
        """

        if not isinstance(other, Nuclide):
            return NotImplemented
        return not self.__eq__(other)

    def __hash__(self) -> int:
        """
        Hash function for ``Nuclide`` instances.
        """

        return hash((self.nuclide, self.decay_data.dataset_name))


def _build_decay_digraph(
    parent: Nuclide,
    digraph: nx.classes.digraph.DiGraph,
) -> nx.classes.digraph.DiGraph:
    """
    Build a networkx DiGraph for the decay chain of this nuclide.

    Parameters
    ----------
    parent : Nuclide
        Nuclide instance of the parent nuclide of the decay chain.
    digraph : networkx.classes.digraph.DiGraph
        DiGraph for the decay chain.

    Returns
    -------
    digraph : networkx.classes.digraph.DiGraph
        DiGraph of the decay chain.
    max_generation : int
        Number of generations of progeny in the decay chain.
    max_xpos : int
        Maximum number of progeny within any one generation of the decay chain.

    """

    generation_max_xpos = {0: 0}

    dequeue = deque([parent.nuclide])
    generations = deque([0])
    xpositions = deque([0])
    node_label = (
        f"{_parse_nuclide_label(parent.nuclide)}\n{parent.half_life('readable')}"
    )
    digraph.add_node(parent.nuclide, generation=0, xpos=0, label=node_label)
    seen = {parent.nuclide}

    while len(dequeue) > 0:
        parent_name = dequeue.popleft()
        generation = generations.popleft() + 1
        xpos = xpositions.popleft()
        if generation not in generation_max_xpos:
            generation_max_xpos[generation] = -1
        parent = Nuclide(parent_name, parent.decay_data)

        progeny = parent.progeny()
        branching_fractions = parent.branching_fractions()
        decay_modes = parent.decay_modes()

        if xpos < generation_max_xpos[generation] + 1:
            xpos = generation_max_xpos[generation] + 1
        xcounter = 0
        for idx, prog in enumerate(progeny):
            if prog not in seen:
                node_label = _parse_nuclide_label(prog)
                if prog in parent.decay_data.nuclide_dict:
                    node_label += f"\n{parent.decay_data.half_life(prog, 'readable')}"
                    if np.isfinite(parent.decay_data.half_life(prog)):
                        dequeue.append(prog)
                        generations.append(generation)
                        xpositions.append(xpos + xcounter)
                if prog == "SF":
                    prog = parent.nuclide + "_SF"

                digraph.add_node(
                    prog,
                    generation=generation,
                    xpos=xpos + xcounter,
                    label=node_label,
                )
                seen.add(prog)

                if xpos + xcounter > generation_max_xpos[generation]:
                    generation_max_xpos[generation] = xpos + xcounter
                xcounter += 1

            edge_label = (
                _parse_decay_mode_label(decay_modes[idx])
                + "\n"
                + str(branching_fractions[idx])
            )
            digraph.add_edge(parent.nuclide, prog, label=edge_label)

    for node in digraph:
        digraph.nodes[node]["pos"] = (
            digraph.nodes[node]["xpos"],
            digraph.nodes[node]["generation"] * -1,
        )

    return digraph, max(generation_max_xpos), max(generation_max_xpos.values())



"""
``radioactivedecay`` is a Python package for radioactive decay calculations.
It supports decay chains of radionuclides, metastable states and branching
decays. By default it uses the decay data from ICRP Publication 107, which
contains 1252 radionuclides of 97 elements, and atomic mass data from the
Atomic Mass Data Center.

The code solves the radioactive decay differential equations analytically using
NumPy and SciPy linear algebra routines. There is also a high numerical
precision calculation mode employing SymPy routines. This gives more accurate
results for decay chains containing radionuclides with orders of magnitude
differences between the half-lives.

This is free-to-use open source software. It was created for engineers,
technicians and researchers who work with radioactivity, and for
educational use.

The docstring code examples assume the ``radioactivedecay`` package has been
imported as ``rd``:

.. highlight:: python
.. code-block:: python

    >>> import radioactivedecay as rd

"""

__version__ = "0.4.19"

from radioactivedecay.decaydata import DEFAULTDATA, DecayData
from radioactivedecay.inventory import Inventory, InventoryHP
from radioactivedecay.nuclide import Nuclide



"""
Unit tests for nuclide.py functions, classes and methods.
"""

import unittest

from radioactivedecay.decaydata import load_dataset
from radioactivedecay.nuclide import Nuclide


class TestNuclide(unittest.TestCase):
    """
    Unit tests for the nuclide.py Nuclide class.
    """

    def test_nuclide_instantiation(self) -> None:
        """
        Test instantiation of Nuclide objects.
        """
        decay_data = load_dataset("icrp107_ame2020_nubase2020", load_sympy=True)

        nuc = Nuclide("Rn-222")
        self.assertEqual(nuc.nuclide, "Rn-222")
        self.assertEqual(nuc.decay_data, decay_data)

        nuc = Nuclide("222Rn")
        self.assertEqual(nuc.nuclide, "Rn-222")

        nuc = Nuclide(611450000)
        self.assertEqual(nuc.nuclide, "Pm-145")

    def test_nuclide_Z(self) -> None:
        """
        Test Nuclide Z property.
        """

        nuc = Nuclide("H-3")
        self.assertEqual(nuc.Z, 1)

    def test_nuclide_A(self) -> None:
        """
        Test Nuclide A property.
        """

        nuc = Nuclide("H-3")
        self.assertEqual(nuc.A, 3)

    def test_nuclide_state(self) -> None:
        """
        Test Nuclide state property.
        """

        nuc = Nuclide("H-3")
        self.assertEqual(nuc.state, "")

        nuc = Nuclide("Ba-137m")
        self.assertEqual(nuc.state, "m")

        nuc = Nuclide("Ir-192n")
        self.assertEqual(nuc.state, "n")

    def test_nuclide_id(self) -> None:
        """
        Test Nuclide id property.
        """

        nuc = Nuclide("H-3")
        self.assertEqual(nuc.id, 10030000)

        nuc = Nuclide("I-118m")
        self.assertEqual(nuc.id, 531180001)

        nuc = Nuclide(190400000)
        self.assertEqual(nuc.id, 190400000)

    def test_nuclide_atomic_mass(self) -> None:
        """
        Test Nuclide atomic_mass property.
        """

        nuc = Nuclide("H-3")
        self.assertEqual(nuc.atomic_mass, 3.01604928132)

    def test_nuclide_half_life(self) -> None:
        """
        Test Nuclide half_life() method.
        """

        nuc = Nuclide("H-3")
        self.assertEqual(nuc.half_life(), 388781329.30560005)
        self.assertEqual(nuc.half_life("y"), 12.32)
        self.assertEqual(nuc.half_life("readable"), "12.32 y")

    def test_nuclide_progeny(self) -> None:
        """
        Test Nuclide progeny() method.
        """

        nuc = Nuclide("K-40")
        self.assertEqual(nuc.progeny()[0], "Ca-40")
        self.assertEqual(nuc.progeny()[1], "Ar-40")

    def test_nuclide_branching_fractions(self) -> None:
        """
        Test Nuclide branching_fractions() method.
        """

        nuc = Nuclide("K-40")
        self.assertEqual(nuc.branching_fractions()[0], 0.8914)
        self.assertEqual(nuc.branching_fractions()[1], 0.1086)

    def test_nuclide_decay_modes(self) -> None:
        """
        Test Nuclide decay_modes() method.
        """

        nuc = Nuclide("K-40")
        self.assertEqual(nuc.decay_modes()[0], "\u03b2-")
        self.assertEqual(nuc.decay_modes()[1], "\u03b2+ & EC")

    def test_nuclide_plot(self) -> None:
        """
        Test Nuclide plot() method.

        Only testing auto-generation of limits so far.
        """

        nuc = Nuclide("H-3")
        _, axes = nuc.plot()
        self.assertEqual(axes.get_xlim(), (-0.3, 0.3))
        self.assertEqual(axes.get_ylim(), (-1.3, 0.3))

        nuc = Nuclide("Mo-99")
        _, axes = nuc.plot()
        self.assertEqual(axes.get_xlim(), (-0.3, 1.3))
        self.assertEqual(axes.get_ylim(), (-2.3, 0.3))

        nuc = Nuclide("Es-256")
        _, axes = nuc.plot()
        self.assertEqual(axes.get_xlim(), (-0.3, 2.3))
        self.assertEqual(axes.get_ylim(), (-19.3, 0.3))

        nuc = Nuclide("Cu-64")
        _, axes = nuc.plot()
        self.assertEqual(axes.get_xlim(), (-0.3, 1.3))
        self.assertEqual(axes.get_ylim(), (-1.3, 0.3))

    def test_nuclide___repr__(self) -> None:
        """
        Test Nuclide __repr__ strings.
        """

        nuc = Nuclide("H-3")
        self.assertEqual(
            repr(nuc),
            "Nuclide: H-3, decay dataset: icrp107_ame2020_nubase2020",
        )

    def test_nuclide___eq__(self) -> None:
        """
        Test Nuclide equality.
        """

        nuc1 = Nuclide("K-40")
        nuc2 = Nuclide("40K")
        nuc3 = Nuclide(190400000)
        self.assertEqual(nuc1, nuc2)
        self.assertEqual(nuc1, nuc3)

        decay_data = load_dataset("icrp107_ame2020_nubase2020", load_sympy=True)
        nuc2 = Nuclide("K-40", decay_data)
        self.assertEqual(nuc1, nuc2)

        self.assertFalse(nuc1 == "random object")

    def test_nuclide___ne__(self) -> None:
        """
        Test Nuclide inequality.
        """

        nuc1 = Nuclide("K-40")
        nuc2 = Nuclide("H-3")
        self.assertNotEqual(nuc1, nuc2)

        self.assertTrue(nuc1 != "random object")

    def test_nuclide___hash__(self) -> None:
        """
        Test Nuclide hash function.
        """

        nuc = Nuclide("K-40")
        decay_data = load_dataset("icrp107_ame2020_nubase2020", load_sympy=True)
        self.assertEqual(hash(nuc), hash(("K-40", decay_data.dataset_name)))


if __name__ == "__main__":
    unittest.main()



"""
The converters module contains classes for unit and quantity conversions.

The docstring code examples assume that ``radioactivedecay`` has been imported
as ``rd``:

.. highlight:: python
.. code-block:: python

    >>> import radioactivedecay as rd

"""

from abc import ABC
from typing import Dict, Union

from sympy import Integer, nsimplify
from sympy.core.expr import Expr

AVOGADRO = 6.02214076e23


class UnitConverter(ABC):
    """
    Template class for unit converters using either floats or SymPy arithmetic.


    Class Attributes
    ----------
    time_units : dict
        Dictionary containing numbers of each time unit per second.
    activity_units : dict
        Dictionary containing amounts of each activity unit per Bq.
    mass_units : dict
        Dictionary containing amounts of each mass unit per g.
    moles_units : dict
        Dictionary containing amounts of each mole unit per mol.
    year_units : set
        Set containing all year units. Used to pick up where days in year conversion is needed.
    time_unit_err_msg : str
        Error message for time unit ValueError.
    activity_unit_err_msg : str
        Error message for activity unit ValueError.
    mass_unit_err_msg : str
        Error message for mass unit ValueError.
    moles_unit_err_msg : str
        Error message for moles unit ValueError.

    """

    time_units: Dict[str, Union[float, Expr]]
    activity_units: Dict[str, Union[float, Expr]]
    mass_units: Dict[str, Union[float, Expr]]
    moles_units: Dict[str, Union[float, Expr]]
    year_units = {"y", "yr", "year", "years", "ky", "My", "By", "Gy", "Ty", "Py"}

    time_unit_err_msg = 'is not a valid time unit, e.g. "s", "m", "h", "d" or "y".'
    activity_unit_err_msg = 'is not a valid activitiy unit, e.g. "Bq", "kBq", "Ci"...'
    mass_unit_err_msg = 'is not a valid mass unit, e.g. "g", "kg", "mg"...'
    moles_unit_err_msg = 'is not a valid moles unit, e.g. "mol", "kmol", "mmol"...'

    @classmethod
    def time_unit_conv(
        cls,
        time_period: Union[float, Expr],
        units_from: str,
        units_to: str,
        year_conv: Union[float, Expr],
    ) -> Union[float, Expr]:
        """
        Converts a time period from one time unit to another.

        Parameters
        ----------
        time_period : float or Expr
            Time period before conversion.
        units_from : str
            Time unit before conversion.
        units_to : str
            Time unit after conversion.
        year_conv : Union[float, Expr]
            Number of days in one year.

        Returns
        -------
        float or Expr
            Time period in new units.

        Raises
        ------
        ValueError
            If one of the time unit parameters is invalid.

        """

        if units_from not in cls.time_units:
            raise ValueError(f"{units_from} {cls.time_unit_err_msg}")
        if units_to not in cls.time_units:
            raise ValueError(f"{units_to} {cls.time_unit_err_msg}")

        factor_from = cls.time_units[units_from]
        factor_to = cls.time_units[units_to]
        if units_from in cls.year_units:
            factor_from *= year_conv
        if units_to in cls.year_units:
            factor_to *= year_conv
        return time_period * factor_from / factor_to

    @classmethod
    def activity_unit_conv(
        cls, activity: Union[float, Expr], units_from: str, units_to: str
    ) -> Union[float, Expr]:
        """
        Converts an activity from one unit to another.

        Parameters
        ----------
        activity : float or Expr
            Activity before conversion.
        units_from : str
            Activity unit before conversion.
        units_to : str
            Activity unit after conversion.

        Returns
        -------
        float or Expr
            Activity in new units.

        Raises
        ------
        ValueError
            If one of the activity unit parameters is invalid.

        """

        if units_from not in cls.activity_units:
            raise ValueError(f"{units_from} {cls.activity_unit_err_msg}")
        if units_to not in cls.activity_units:
            raise ValueError(f"{units_to} {cls.activity_unit_err_msg}")

        return activity * cls.activity_units[units_from] / cls.activity_units[units_to]

    @classmethod
    def mass_unit_conv(
        cls, mass: Union[float, Expr], units_from: str, units_to: str
    ) -> Union[float, Expr]:
        """
        Converts a mass from one unit to another.

        Parameters
        ----------
        mass : float or Expr
            Mass before conversion.
        units_from : str
            Mass unit before conversion.
        units_to : str
            Mass unit after conversion.

        Returns
        -------
        float or Expr
            Mass in new units.

        Raises
        ------
        ValueError
            If one of the mass unit parameters is invalid.

        """

        if units_from not in cls.mass_units:
            raise ValueError(f"{units_from} {cls.mass_unit_err_msg}")
        if units_to not in cls.mass_units:
            raise ValueError(f"{units_to} {cls.mass_unit_err_msg}")

        return mass * cls.mass_units[units_from] / cls.mass_units[units_to]

    @classmethod
    def moles_unit_conv(
        cls, moles: Union[float, Expr], units_from: str, units_to: str
    ) -> Union[float, Expr]:
        """
        Converts a number of moles from one order of magnitude to another.

        Parameters
        ----------
        moles : float or Expr
            Amount before conversion.
        units_from : str
            Unit before conversion.
        units_to : str
            Unit after conversion.

        Returns
        -------
        float or Expr
            Result in chosen units.

        Raises
        ------
        ValueError
            If one of the unit parameters is invalid.

        """

        if units_from not in cls.moles_units:
            raise ValueError(f"{units_from} {cls.moles_unit_err_msg}")
        if units_to not in cls.moles_units:
            raise ValueError(f"{units_to} {cls.moles_unit_err_msg}")

        return moles * cls.moles_units[units_from] / cls.moles_units[units_to]


class UnitConverterFloat(UnitConverter):
    """
    Unit converter using floats.
    """

    time_units: Dict[str, float] = {
        "ps": 1.0e-12,
        "ns": 1.0e-9,
        "μs": 1.0e-6,
        "us": 1.0e-6,
        "ms": 1.0e-3,
        "s": 1.0,
        "m": 60.0,
        "h": 3600.0,
        "d": 86400.0,
        "y": 86400.0,
        "sec": 1,
        "second": 1,
        "seconds": 1,
        "hr": 3600.0,
        "hour": 3600.0,
        "hours": 3600.0,
        "day": 86400.0,
        "days": 86400.0,
        "yr": 86400.0,
        "year": 86400.0,
        "years": 86400.0,
        "ky": 86400.0 * 1.0e3,
        "My": 86400.0 * 1.0e6,
        "By": 86400.0 * 1.0e9,
        "Gy": 86400.0 * 1.0e9,
        "Ty": 86400.0 * 1.0e12,
        "Py": 86400.0 * 1.0e15,
    }

    activity_units: Dict[str, float] = {
        "pBq": 1.0e-12,
        "nBq": 1.0e-9,
        "μBq": 1.0e-6,
        "uBq": 1.0e-6,
        "mBq": 1.0e-3,
        "Bq": 1.0,
        "kBq": 1.0e3,
        "MBq": 1.0e6,
        "GBq": 1.0e9,
        "TBq": 1.0e12,
        "PBq": 1.0e15,
        "EBq": 1.0e18,
        "pCi": 1.0e-12 * 3.7e10,
        "nCi": 1.0e-9 * 3.7e10,
        "μCi": 1.0e-6 * 3.7e10,
        "uCi": 1.0e-6 * 3.7e10,
        "mCi": 1.0e-3 * 3.7e10,
        "Ci": 1.0 * 3.7e10,
        "kCi": 1.0e3 * 3.7e10,
        "MCi": 1.0e6 * 3.7e10,
        "GCi": 1.0e9 * 3.7e10,
        "TCi": 1.0e12 * 3.7e10,
        "PCi": 1.0e15 * 3.7e10,
        "ECi": 1.0e18 * 3.7e10,
        "dpm": 1.0 / 60.0,
    }

    mass_units: Dict[str, float] = {
        "pg": 1.0e-12,
        "ng": 1.0e-9,
        "μg": 1.0e-6,
        "ug": 1.0e-6,
        "mg": 1.0e-3,
        "g": 1.0,
        "kg": 1.0e3,
        "Mg": 1.0e6,
        "t": 1.0e6,
        "ton": 1.0e6,
    }

    moles_units: Dict[str, float] = {
        "pmol": 1.0e-12,
        "nmol": 1.0e-9,
        "μmol": 1.0e-6,
        "umol": 1.0e-6,
        "mmol": 1.0e-3,
        "mol": 1.0,
        "kmol": 1.0e3,
        "Mmol": 1.0e6,
    }

    @classmethod
    def __repr__(cls) -> str:
        return f"{cls.__name__} using double-precision floats."


class UnitConverterSympy(UnitConverter):
    """
    Unit converter using SymPy arbitrary precision operations.
    """

    time_units: Dict[str, Expr] = {
        "ps": Integer(1) / 1000000000000,
        "ns": Integer(1) / 1000000000,
        "μs": Integer(1) / 1000000,
        "us": Integer(1) / 1000000,
        "ms": Integer(1) / 1000,
        "s": Integer(1),
        "m": Integer(60),
        "h": Integer(3600),
        "d": Integer(86400),
        "y": Integer(86400),
        "sec": Integer(1),
        "second": Integer(1),
        "seconds": Integer(1),
        "hr": Integer(3600),
        "hour": Integer(3600),
        "hours": Integer(3600),
        "day": Integer(86400),
        "days": Integer(86400),
        "yr": Integer(86400),
        "year": Integer(86400),
        "years": Integer(86400),
        "ky": Integer(86400) * 1000,
        "My": Integer(86400) * 1000000,
        "By": Integer(86400) * 1000000000,
        "Gy": Integer(86400) * 1000000000,
        "Ty": Integer(86400) * 1000000000000,
        "Py": Integer(86400) * 1000000000000000,
    }

    activity_units: Dict[str, Expr] = {
        "pBq": Integer(1) / 1000000000000,
        "nBq": Integer(1) / 1000000000,
        "μBq": Integer(1) / 1000000,
        "uBq": Integer(1) / 1000000,
        "mBq": Integer(1) / 1000,
        "Bq": Integer(1),
        "kBq": Integer(1000),
        "MBq": Integer(1000000),
        "GBq": Integer(1000000000),
        "TBq": Integer(1000000000000),
        "PBq": Integer(1000000000000000),
        "EBq": Integer(1000000000000000000),
        "pCi": Integer(1) / 1000000000000 * 37000000000,
        "nCi": Integer(1) / 1000000000 * 37000000000,
        "μCi": Integer(1) / 1000000 * 37000000000,
        "uCi": Integer(1) / 1000000 * 37000000000,
        "mCi": Integer(1) / 1000 * 37000000000,
        "Ci": Integer(1) * 37000000000,
        "kCi": Integer(1000) * 37000000000,
        "MCi": Integer(1000000) * 37000000000,
        "GCi": Integer(1000000000) * 37000000000,
        "TCi": Integer(1000000000000) * 37000000000,
        "PCi": Integer(1000000000000000) * 37000000000,
        "ECi": Integer(1000000000000000000) * 37000000000,
        "dpm": Integer(1) / 60,
    }

    mass_units: Dict[str, Expr] = {
        "pg": Integer(1) / 1000000000000,
        "ng": Integer(1) / 1000000000,
        "μg": Integer(1) / 1000000,
        "ug": Integer(1) / 1000000,
        "mg": Integer(1) / 1000,
        "g": Integer(1),
        "kg": Integer(1000),
        "Mg": Integer(1000000),
        "t": Integer(1000000),
        "ton": Integer(1000000),
    }

    moles_units: Dict[str, Expr] = {
        "pmol": Integer(1) / 1000000000000,
        "nmol": Integer(1) / 1000000000,
        "μmol": Integer(1) / 1000000,
        "umol": Integer(1) / 1000000,
        "mmol": Integer(1) / 1000,
        "mol": Integer(1),
        "kmol": Integer(1000),
        "Mmol": Integer(1000000),
    }

    @classmethod
    def __repr__(cls) -> str:
        return f"{cls.__name__} using SymPy arbitrary precision calculations."


class QuantityConverter(ABC):
    """
    Template class for quantity converters using either floats or SymPy arithmetic. Converts
    activity in Bq, mass in g and moles in mol to number of atoms, and vice versa.

    Class Attributes
    ----------
    avogadro : float or Expr
        Avogadro constant (number of atoms/mol).

    """

    avogadro: Union[float, Expr]

    @staticmethod
    def activity_to_number(
        activity: Union[float, Expr], decay_const: Union[float, Expr]
    ) -> Union[float, Expr]:
        """
        Converts an activity in Bq to the number of atoms.

        Parameters
        ----------
        activity : float or sympy.core.expr.Expr
            The activity in Bq of the nuclide to be converted.
        decay_const : float or sympy.core.expr.Expr
            Decay constant of the nuclide in s^-1.

        Returns
        -------
        float or sympy.core.expr.Expr
            Number of atoms of the nuclide.

        """

        return activity / decay_const

    @classmethod
    def mass_to_number(
        cls, mass: Union[float, Expr], atomic_mass: Union[float, Expr]
    ) -> Union[float, Expr]:
        """
        Converts a mass in grams to number of atoms.

        Parameters
        ----------
        mass : float or sympy.core.expr.Expr
            The mass of the nuclide to be converted in grams.
        atomic_mass : float or sympy.core.expr.Expr
            Atomic mass of the nuclide in g/mol.

        Returns
        -------
        float or sympy.core.expr.Expr
            Number of atoms of the nuclide.

        """

        return mass / atomic_mass * cls.avogadro

    @classmethod
    def moles_to_number(cls, moles: Union[float, Expr]) -> Union[float, Expr]:
        """
        Converts number of moles to number of atoms.

        Parameters
        ----------
        moles : float or sympy.core.expr.Expr
            Number of moles to be converted.

        Returns
        -------
        float or sympy.core.expr.Expr
            Number of atoms.

        """

        return moles * cls.avogadro

    @staticmethod
    def number_to_activity(
        number: Union[float, Expr], decay_const: Union[float, Expr]
    ) -> Union[float, Expr]:
        """
        Converts number of atoms to activity in Bq.

        Parameters
        ----------
        number : float or sympy.core.expr.Expr
            The number of atoms of nuclide to be converted.
        decay_const : float or sympy.core.expr.Expr
            Decay constant of the nuclide in s^-1.

        Returns
        -------
        float or sympy.core.expr.Expr
            Activity of the nuclide in Bq.

        """

        return number * decay_const

    @classmethod
    def number_to_mass(
        cls, number: Union[float, Expr], atomic_mass: Union[float, Expr]
    ) -> Union[float, Expr]:
        """
        Converts number of atoms to mass in grams. Supports both Exprs or SymPy quantities.

        Parameters
        ----------
        number : float or sympy.core.expr.Expr
            The number of atoms of the nuclide to be converted.
        atomic_mass : float or sympy.core.expr.Expr
            Atomic mass of the nuclide in g/mol.

        Returns
        -------
        float or sympy.core.expr.Expr
            Mass of material in grams.

        """

        return number / cls.avogadro * atomic_mass

    @classmethod
    def number_to_moles(cls, number: Union[float, Expr]) -> Union[float, Expr]:
        """
        Converts number of atoms to moles of nuclide.

        Parameters
        ----------
        number : float or sympy.core.expr.Expr
            The number of atoms of the nuclide to be converted.

        Returns
        -------
        float or sympy.core.expr.Expr
            Moles of nuclide.

        """

        return number / cls.avogadro


class QuantityConverterFloat(QuantityConverter):
    """
    Quantity converter using SymPy arbitrary precision operations.
    """

    avogadro: float = AVOGADRO

    @classmethod
    def __repr__(cls) -> str:
        return f"{cls.__name__} using double-precision floats."


class QuantityConverterSympy(QuantityConverter):
    """
    Quantity converter using SymPy arbitrary precision operations.
    """

    avogadro: Expr = nsimplify(AVOGADRO)

    @classmethod
    def __repr__(cls) -> str:
        return f"{cls.__name__} using SymPy arbitrary precision calculations."



"""
Unit tests for inventory.py classes and methods.
"""

import copy
import math
import unittest
import warnings
from typing import Dict
from unittest.mock import patch

import numpy as np
from sympy import Integer, log

from radioactivedecay.decaydata import DEFAULTDATA, load_dataset
from radioactivedecay.inventory import Inventory, InventoryHP
from radioactivedecay.nuclide import Nuclide

# pylint: disable=protected-access, too-many-public-methods


def warning_message_if_dict_not_equal(
    calculated: Dict[str, float], expected: Dict[str, float]
) -> None:
    """
    Warning message if calculated dictionary of floats is not equal to expected dictionary of
    floats.
    """

    return (
        f"Warning: calculated result:\n{calculated}\nis not identical to expected result:\n"
        f"{expected}\nFloating point error build up or other code issue has occurred."
    )


def dict_assert_almost_equal(
    test_case: unittest.TestCase,
    dict_a: Dict[str, float],
    dict_b: Dict[str, float],
) -> None:
    """
    Check whether two dictionaries with str keys and float values have:
        i) the same keys
        ii) float values are close (using `math.isclose()`)

    This function could be improved by outputting a better traceback in the event that an
    assertTrue fails.
    """

    test_case.assertEqual(set(dict_a), set(dict_b))

    for key in dict_a:
        try:
            test_case.assertTrue(
                math.isclose(dict_a[key], dict_b[key], rel_tol=1e-7, abs_tol=1e-30)
            )
        except AssertionError as error:
            raise AssertionError(
                f"Floats are not within error tolerance: {dict_a[key]} {dict_b[key]} for nuclide "
                f"{key}."
            ) from error


class TestInventory(unittest.TestCase):
    """
    Unit tests for the inventory.py Inventory class.
    """

    def test_instantiation(self) -> None:
        """
        Test instantiation of Inventory objects.
        """

        # check basic instantiation
        inv = Inventory({"H3": 1}, "num", True)
        self.assertEqual(inv.contents, {"H-3": 1})
        self.assertEqual(inv.decay_data, DEFAULTDATA)
        self.assertEqual(inv.decay_matrices, DEFAULTDATA.scipy_data)

        # check instantiation using Nuclide instance
        tritium = Nuclide("H3")
        inv = Inventory({tritium: 1.0}, "num", True)
        self.assertEqual(inv.contents, {"H-3": 1})

        # check that check=False turns off nuclide string conversion and quantity check
        inv = Inventory({"H3": 1.0}, "num", False)
        self.assertEqual(inv.contents, {"H3": 1.0})
        inv = Inventory({"H-3": -1.0}, "num", False)
        self.assertEqual(inv.contents, {"H-3": -1.0})

        # check sorting of contents dictionary
        inv = Inventory({"He-3": 2.0, "H-3": 1.0}, "num", False)
        self.assertEqual(inv.contents, {"H-3": 1.0, "He-3": 2.0})

        # check conversion of input quantity to number
        inv = Inventory({"H-3": 1.0}, "mBq")
        self.assertEqual(inv.contents, {"H-3": 560892.8957794083})

        # check instantiation with a stable nuclide activity raises a ValueError
        with self.assertRaises(ValueError):
            Inventory({"He-3": 0.0}, "Bq", False)

        # check instantiation with NumPy int datatype
        inv = Inventory({"H-3": np.int32(1)}, "num")
        self.assertEqual(inv.contents, {"H-3": 1})

    def test__parse_nuclides(self) -> None:
        """
        Test the conversion of nuclide strings or Nuclide instances in a contents dictionary
        to nuclide strings into Ab-XY form.
        """

        nuclides = DEFAULTDATA.nuclides
        dataset_name = DEFAULTDATA.dataset_name
        self.assertEqual(
            Inventory._parse_nuclides({"H-3": 1.0}, nuclides, dataset_name),
            {"H-3": 1.0},
        )
        self.assertEqual(
            Inventory._parse_nuclides({"He3": 1.0}, nuclides, dataset_name),
            {"He-3": 1.0},
        )

        tritium = Nuclide("H-3")
        self.assertEqual(
            Inventory._parse_nuclides(
                {tritium: 1.0, "3He": 2.0}, nuclides, dataset_name
            ),
            {"H-3": 1.0, "He-3": 2.0},
        )

    def test__check_values(self) -> None:
        """
        Test that the contents dictionary values (amounts of each nuclide) are physical.
        """

        self.assertIsNone(Inventory._check_values({"H-3": 1.0}))
        self.assertIsNone(Inventory._check_values({"H-3": 0}))
        self.assertIsNone(Inventory._check_values({"H-3": Integer(0)}))

        with self.assertRaises(ValueError):
            Inventory._check_values({"H-3": -1})
        with self.assertRaises(ValueError):
            Inventory._check_values({"H-3": "1"})

    def test__convert_to_number(self) -> None:
        """
        Test that the contents dictionary values (amounts of each nuclide) are physical.
        """

        inv = Inventory({"H3": 1}, "num", True)
        self.assertEqual(
            inv._convert_to_number({"H-3": 1.0}, "num", "dataset_name"),
            {"H-3": 1.0},
        )
        self.assertEqual(
            inv._convert_to_number({"H-3": 1.0}, "Bq", "dataset_name"),
            {"H-3": 560892895.7794082},
        )
        self.assertEqual(
            inv._convert_to_number({"H-3": 1.0}, "mBq", "dataset_name"),
            {"H-3": 560892.8957794083},
        )
        self.assertEqual(
            inv._convert_to_number({"H-3": 1.0}, "mol", "dataset_name"),
            {"H-3": 6.02214076e23},
        )
        self.assertEqual(
            inv._convert_to_number({"He-3": 1.0}, "kg", "dataset_name"),
            {"He-3": 1.9967116089131648e26},
        )

        # check catch of incorrect units
        with self.assertRaises(ValueError):
            inv._convert_to_number({"H-3": 1.0}, "xyz", "dataset_name")

    def test_nuclides(self) -> None:
        """
        Test Inventory.nuclides property.
        """

        inv = Inventory({"H-3": 1.0})
        self.assertEqual(inv.nuclides, ["H-3"])
        inv = Inventory({"Tc-99m": 2.3, "I-123": 5.8})
        self.assertEqual(inv.nuclides, ["I-123", "Tc-99m"])

    def test_numbers(self) -> None:
        """
        Test Inventory.numbers() method.
        """

        inv = Inventory({"H-3": 1}, "num")
        self.assertEqual(inv.numbers(), {"H-3": 1})
        inv = Inventory({"Tc-99m": 2.3, "I-123": 5.8})
        self.assertEqual(
            inv.numbers(), {"I-123": 399738.47946141585, "Tc-99m": 71852.27235544211}
        )

    def test_activities(self) -> None:
        """
        Test Inventory.activities() method.
        """

        inv = Inventory({"H-3": 1})
        self.assertEqual(inv.activities(), {"H-3": 1})
        inv = Inventory({"Tc-99m": 2.3, "I-123": 5.8})
        self.assertEqual(inv.activities(), {"I-123": 5.8, "Tc-99m": 2.3})
        inv = Inventory({"Tc-99m": 2300, "I-123": 5800})
        self.assertEqual(inv.activities("kBq"), {"I-123": 5.8, "Tc-99m": 2.3})

    def test_masses(self) -> None:
        """
        Test Inventory.masses() method.
        """

        inv = Inventory({"H-3": 1}, "g")
        self.assertEqual(inv.masses(), {"H-3": 1})
        inv = Inventory({"Tc-99m": 2.3, "I-123": 5.8})
        self.assertEqual(
            inv.masses(),
            {"I-123": 8.158243973887584e-17, "Tc-99m": 1.1800869622748501e-17},
        )
        self.assertEqual(
            inv.masses("pg"),
            {"I-123": 8.158243973887584e-05, "Tc-99m": 1.1800869622748502e-05},
        )

    def test_moles(self) -> None:
        """
        Test Inventory.moles() method.
        """

        inv = Inventory({"H-3": 1}, "mol")
        self.assertEqual(inv.moles(), {"H-3": 1})
        inv = Inventory({"Tc-99m": 2.3, "I-123": 5.8})
        self.assertEqual(
            inv.moles(),
            {"I-123": 6.637813617983513e-19, "Tc-99m": 1.1931350531142702e-19},
        )
        self.assertEqual(
            inv.moles("mmol"),
            {"I-123": 6.637813617983513e-16, "Tc-99m": 1.1931350531142702e-16},
        )

    def test_activity_fractions(self) -> None:
        """
        Test Inventory.activity_fractions() method.
        """

        inv = Inventory({"Tc-99m": 2.3, "I-123": 5.8})
        self.assertEqual(
            inv.activity_fractions(),
            {"I-123": 0.7160493827160493, "Tc-99m": 0.2839506172839506},
        )
        self.assertEqual(sum(inv.activity_fractions().values()), 1.0)

    def test_mass_fractions(self) -> None:
        """
        Test Inventory.mass_fractions() method.
        """

        inv = Inventory({"Tc-99m": 2.3, "I-123": 5.8})
        self.assertEqual(
            inv.mass_fractions(),
            {"I-123": 0.8736297770616593, "Tc-99m": 0.12637022293834066},
        )
        self.assertEqual(sum(inv.mass_fractions().values()), 1.0)

    def test_mole_fractions(self) -> None:
        """
        Test Inventory.mole_fractions() method.
        """

        inv = Inventory({"Tc-99m": 2.3, "I-123": 5.8})
        self.assertEqual(
            inv.mole_fractions(),
            {"I-123": 0.8476385041932588, "Tc-99m": 0.15236149580674116},
        )
        self.assertEqual(sum(inv.mole_fractions().values()), 1.0)

    def test___len__(self) -> None:
        """
        Test len() on Inventory.
        """

        inv = Inventory({"H-3": 1})
        self.assertEqual(len(inv), 1)
        inv = Inventory({"Tc-99m": 2.3, "I-123": 5.8})
        self.assertEqual(len(inv), 2)

    def test_add(self) -> None:
        """
        Test Inventory.add() method to append to an inventory.
        """

        inv = Inventory({"H-3": 1}, "num")
        inv.add({60140000: 3.0, "K-40": 4.0}, "num")
        self.assertEqual(inv.contents, {"C-14": 3.0, "H-3": 1.0, "K-40": 4.0})
        inv.add({"H-3": 3.0}, "num")
        self.assertEqual(inv.contents, {"C-14": 3.0, "H-3": 4.0, "K-40": 4.0})

        inv = Inventory({"H-3": 1}, "num")
        inv.add({Nuclide("C-14"): 3.0, "K-40": 4.0}, "num")
        self.assertEqual(inv.contents, {"C-14": 3.0, "H-3": 1.0, "K-40": 4.0})
        inv.add({Nuclide("H-3"): 3.0}, "num")
        self.assertEqual(inv.contents, {"C-14": 3.0, "H-3": 4.0, "K-40": 4.0})

    def test_subtract(self) -> None:
        """
        Test Inventory.subtract() method to take away a dictionary from an inventory.
        """

        inv = Inventory({"C-14": 3.0, "H-3": 4.0, "K-40": 4.0}, "num")
        inv.subtract({60140000: 3.0, "K-40": 4.0}, "num")
        self.assertEqual(inv.contents, {"C-14": 0.0, "H-3": 4.0, "K-40": 0.0}, "num")

        inv = Inventory({"C-14": 3.0, "H-3": 4.0, "K-40": 4.0}, "num")
        inv.subtract({"C-14": 3.0, Nuclide("K-40"): 4.0}, "num")
        self.assertEqual(inv.contents, {"C-14": 0.0, "H-3": 4.0, "K-40": 0.0}, "num")

    def test___add__(self) -> None:
        """
        Test operator to add two inventory objects together.
        """

        inv1 = Inventory({"H-3": 1.0}, "num")
        inv2 = Inventory({"C-14": 1.0, "H-3": 4.0}, "num")
        inv = inv1 + inv2
        self.assertEqual(inv.contents, {"C-14": 1.0, "H-3": 5.0}, "num")

        temp_data = copy.deepcopy(DEFAULTDATA)
        temp_data.dataset_name = "icrp107_"
        inv3 = Inventory({"H-3": 2.0}, "num", decay_data=temp_data)
        with self.assertRaises(ValueError):
            inv = inv1 + inv3

    def test___subtract__(self) -> None:
        """
        Test operator to subtract one inventory object from another.
        """

        inv1 = Inventory({"H-3": 1.0}, "num")
        inv2 = Inventory({"C-14": 1.0, "H-3": 4.0}, "num")
        inv = inv2 - inv1
        self.assertEqual(inv.contents, {"C-14": 1.0, "H-3": 3.0})

        temp_data = copy.deepcopy(DEFAULTDATA)
        temp_data.dataset_name = "icrp107_"
        inv3 = Inventory({"H-3": 2.0}, "num", decay_data=temp_data)
        with self.assertRaises(ValueError):
            inv = inv1 - inv3

    def test___mul__(self) -> None:
        """
        Test operator to multiply activities in inventory by constant.
        """

        inv = Inventory({"Sr-90": 1.0, "Cs-137": 1.0}, "num")
        inv = inv * 2
        self.assertEqual(inv.contents, {"Cs-137": 2.0, "Sr-90": 2.0})

    def test___rmul__(self) -> None:
        """
        Test operator to right multiply constant by activities in inventory.
        """

        inv = Inventory({"Sr-90": 1.0, "Cs-137": 1.0}, "num")
        inv = 2 * inv
        self.assertEqual(inv.contents, {"Cs-137": 2.0, "Sr-90": 2.0})

    def test___truediv__(self) -> None:
        """
        Test operator to multiply activities in inventory by constant.
        """

        inv = Inventory({"Sr-90": 1.0, "Cs-137": 1.0}, "num")
        inv = inv / 2
        self.assertEqual(inv.contents, {"Cs-137": 0.5, "Sr-90": 0.5})

    def test_remove(self) -> None:
        """
        Test operator to remove nuclides from an inventory.
        """

        inv = Inventory({"C-14": 3.0, "H-3": 4.0, "K-40": 4.0}, "num")
        with self.assertRaises(NotImplementedError):
            inv.remove(1.0)

    def test_remove_string(self) -> None:
        """
        Test operator to remove one nuclide from an inventory using a nuclide string.
        """

        inv = Inventory({"C-14": 3.0, "H-3": 4.0, "K-40": 4.0}, "num")
        inv.remove("H-3")
        self.assertEqual(inv.contents, {"C-14": 3.0, "K-40": 4.0})

        with self.assertRaises(ValueError):
            inv.remove("Be-10")

    def test_remove_canonical_id(self) -> None:
        """
        Test operator to remove one nuclide from an inventory using a nuclide canonical id.
        """

        inv = Inventory({"C-14": 3.0, "H-3": 4.0, "K-40": 4.0}, "num")
        inv.remove(10030000)
        self.assertEqual(inv.contents, {"C-14": 3.0, "K-40": 4.0})

        with self.assertRaises(ValueError):
            inv.remove(40100000)

    def test_remove_nuclide(self) -> None:
        """
        Test operator to remove one nuclide from an inventory using a ``Nuclide`` object.
        """

        inv = Inventory({"C-14": 3.0, "H-3": 4.0, "K-40": 4.0}, "num")
        inv.remove(Nuclide("H-3"))
        self.assertEqual(inv.contents, {"C-14": 3.0, "K-40": 4.0})

        with self.assertRaises(ValueError):
            inv.remove(Nuclide("Be-10"))

    def test_remove_list(self) -> None:
        """
        Test operator to remove list of nuclides from an inventory.
        """

        inv = Inventory({"C-14": 3.0, "H-3": 4.0, "K-40": 4.0}, "num")
        inv.remove([10030000, "C-14"])
        self.assertEqual(inv.contents, {"K-40": 4.0})

        with self.assertRaises(ValueError):
            inv.remove(["Be-10", "C-14"])

        inv = Inventory({"C-14": 3.0, "H-3": 4.0, "K-40": 4.0}, "num")
        inv.remove(["H-3", Nuclide("C-14")])
        self.assertEqual(inv.contents, {"K-40": 4.0})

    def test_decay(self) -> None:
        """
        Test Inventory.decay() calculations.
        """

        inv = Inventory({"H-3": 10.0}, "Bq")
        self.assertEqual(inv.decay(12.32, "y").activities(), {"H-3": 5.0, "He-3": 0.0})

        inv = Inventory({"Tc-99m": 2.3, "I-123": 5.8}, "Bq")
        calculated = inv.decay(20.0, "h").activities()
        expected = {
            "I-123": 2.040459244534774,
            "Ru-99": 0.0,
            "Sb-123": 0.0,
            "Tc-99": 6.729944738772211e-09,
            "Tc-99m": 0.22950748010063513,
            "Te-123": 9.485166535243877e-18,
            "Te-123m": 7.721174031572363e-07,
        }
        if calculated != expected:
            warnings.warn(warning_message_if_dict_not_equal(calculated, expected))
        dict_assert_almost_equal(self, calculated, expected)

        inv = Inventory({"U-238": 99.274, "U-235": 0.720, "U-234": 0.005}, "Bq")
        calculated = inv.decay(1e9, "y").activities()
        expected = {
            "Ac-227": 0.2690006281740556,
            "At-218": 0.017002868638497183,
            "At-219": 2.227325201281319e-07,
            "Bi-210": 85.01434361515662,
            "Bi-211": 0.26900084425585846,
            "Bi-214": 85.01432618961896,
            "Bi-215": 2.1605054452429237e-07,
            "Fr-223": 0.0037122086688021884,
            "Hg-206": 1.6152725286830197e-06,
            "Pa-231": 0.2690006198549055,
            "Pa-234": 0.13601313171698984,
            "Pa-234m": 85.00820732310412,
            "Pb-206": 0.0,
            "Pb-207": 0.0,
            "Pb-210": 85.01434361489548,
            "Pb-211": 0.2690008442558569,
            "Pb-214": 84.99734032384839,
            "Po-210": 85.01434362236536,
            "Po-211": 0.0007424423301461693,
            "Po-214": 84.99649018398776,
            "Po-215": 0.26900084425583065,
            "Po-218": 85.01434319248591,
            "Ra-223": 0.26900062820528614,
            "Ra-226": 85.01434319228659,
            "Rn-218": 1.7002868638497185e-05,
            "Rn-219": 0.26900062820528614,
            "Rn-222": 85.0143431924858,
            "Th-227": 0.2652884195245263,
            "Th-230": 85.01431274847525,
            "Th-231": 0.26898810215560653,
            "Th-234": 85.00820732310407,
            "Tl-206": 0.00011383420610068998,
            "Tl-207": 0.26825840192571576,
            "Tl-210": 0.01785300849981999,
            "U-234": 85.01287846492669,
            "U-235": 0.2689881021544942,
            "U-238": 85.00820732184867,
        }
        if calculated != expected:
            warnings.warn(warning_message_if_dict_not_equal(calculated, expected))
        dict_assert_almost_equal(self, calculated, expected)

    def test_cumulative_decays(self) -> None:
        """
        Test Inventory.cumulative_decays() calculations.
        """

        inv = Inventory({"H-3": 10.0}, "num")
        self.assertEqual(inv.cumulative_decays(12.32, "y"), {"H-3": 5.0})
        self.assertEqual(inv.cumulative_decays(1e6, "y"), {"H-3": 10.0})

        inv = Inventory({"Sr-90": 10.0}, "num")
        self.assertEqual(
            inv.cumulative_decays(1e6, "y"), {"Sr-90": 10.0, "Y-90": 10.000000000000002}
        )

        inv = Inventory({"Tc-99m": 2.3, "I-123": 5.8}, "num")
        calculated = inv.cumulative_decays(20.0, "h")
        expected = {
            "I-123": 3.759540755465226,
            "Te-123m": 4.72801274418656e-07,
            "Te-123": -9.485843969524933e-18,
            "Tc-99m": 2.0704925198993647,
            "Tc-99": 1.0500063924036559e-08,
        }
        if calculated != expected:
            warnings.warn(warning_message_if_dict_not_equal(calculated, expected))
        dict_assert_almost_equal(self, calculated, expected)

        inv = Inventory({"U-238": 99.274, "U-235": 0.720, "U-234": 0.005}, "num")
        calculated = inv.cumulative_decays(1e9, "y")
        expected = {
            "Ac-227": 0.45099937182594435,
            "Th-227": 0.44477558047547366,
            "Ra-226": 14.264656807713404,
            "Fr-223": 0.006223791331197811,
            "Ra-223": 0.4509993717947137,
            "Rn-222": 14.264656807514221,
            "Rn-219": 0.4509993717947137,
            "At-219": 3.734274798718682e-07,
            "Po-218": 14.264656807514106,
            "At-218": 0.002852931361502821,
            "Rn-218": 2.852931361502822e-06,
            "Bi-215": 3.622246554757077e-07,
            "Po-215": 0.4509997340193692,
            "Pb-214": 14.261803876151637,
            "Bi-214": 14.264653954581055,
            "Po-214": 14.261661230181954,
            "Pb-211": 0.450999734019343,
            "Bi-211": 0.4509997340193414,
            "Po-211": 0.0012447592658933822,
            "Tl-210": 0.0029955773304620116,
            "Pb-210": 14.264656385104534,
            "Bi-210": 14.264656384843379,
            "Po-210": 14.26465637763465,
            "U-238": 14.265792678151335,
            "Tl-207": 0.4497549747534446,
            "Hg-206": 2.710284713169805e-07,
            "Tl-206": 1.910037489931004e-05,
            "U-235": 0.4510118978455057,
            "Th-234": 14.26579267689593,
            "Pa-234m": 14.265792676895886,
            "Pa-234": 0.022825268283010153,
            "U-234": 14.266121535073315,
            "Th-231": 0.45101189784439344,
            "Pa-231": 0.45099938014509444,
            "Th-230": 14.264687251524753,
        }
        if calculated != expected:
            warnings.warn(warning_message_if_dict_not_equal(calculated, expected))
        dict_assert_almost_equal(self, calculated, expected)

    def test_half_lives(self) -> None:
        """
        Test method to fetch half-lives of nuclides in the Inventory.
        """

        inv = Inventory({"C-14": 1.0, "H-3": 2.0}, "num")
        self.assertEqual(inv.half_lives("y"), {"C-14": 5700.0, "H-3": 12.32})
        self.assertEqual(
            inv.half_lives("readable"), {"C-14": "5.70 ky", "H-3": "12.32 y"}
        )

    def test_progeny(self) -> None:
        """
        Test method to fetch progeny of nuclides in the Inventory.
        """

        inv = Inventory({"C-14": 1.0, "K-40": 2.0}, "num")
        self.assertEqual(inv.progeny(), {"C-14": ["N-14"], "K-40": ["Ca-40", "Ar-40"]})

    def test_branching_fractions(self) -> None:
        """
        Test method to fetch branching fractions of nuclides in the Inventory.
        """

        inv = Inventory({"C-14": 1.0, "K-40": 2.0}, "num")
        self.assertEqual(
            inv.branching_fractions(), {"C-14": [1.0], "K-40": [0.8914, 0.1086]}
        )

    def test_decay_modes(self) -> None:
        """
        Test method to fetch decay modes of nuclides in the Inventory.
        """

        inv = Inventory({"C-14": 1.0, "K-40": 2.0}, "num")
        self.assertEqual(
            inv.decay_modes(),
            {"C-14": ["\u03b2-"], "K-40": ["\u03b2-", "\u03b2+ \u0026 EC"]},
        )

    @patch("matplotlib.pyplot.show")
    def test_plot(self, mock_show) -> None:
        """
        Test method to create decay plots.
        """

        inv = Inventory({"C-14": 1.0, "K-40": 2.0}, "MBq")
        _, axes = inv.plot(105, "ky", yunits="kBq")
        self.assertEqual(axes.get_xscale(), "linear")
        self.assertEqual(axes.get_yscale(), "linear")
        self.assertEqual(axes.get_xlabel(), "Time (ky)")
        self.assertEqual(axes.get_ylabel(), "Activity (kBq)")
        self.assertEqual(axes.get_xlim(), (-5.25, 110.25))
        self.assertEqual(axes.get_ylim(), (0.0, 2100.0))
        self.assertEqual(
            axes.get_legend_handles_labels()[-1],
            ["K-40", "Ca-40", "Ar-40", "C-14", "N-14"],
        )

        inv = Inventory({"C-14": 1.0, "K-40": 2.0}, "kg")
        _, axes = inv.plot(
            100, "ky", xmin=50, ymin=1.0, ymax=2.5, yunits="kg", display="K40"
        )
        self.assertEqual(axes.get_xlim(), (47.5, 102.5))
        self.assertEqual(axes.get_ylim(), (1.0, 2.5))
        self.assertEqual(axes.get_ylabel(), "Mass (kg)")
        self.assertEqual(axes.get_legend_handles_labels()[-1], ["K-40"])

        inv = Inventory({"C-14": 1.0, "K-40": 2.0}, "num")
        _, axes = inv.plot(100, "ky", yunits="num", order="alphabetical")
        self.assertEqual(
            axes.get_legend_handles_labels()[-1],
            ["Ar-40", "C-14", "Ca-40", "K-40", "N-14"],
        )
        self.assertEqual(axes.get_ylabel(), "Number of atoms")

        inv = Inventory({"C-14": 1.0, "K-40": 2.0}, "Bq")
        _, axes = inv.plot(100, "ky", yunits="activity_frac")
        self.assertEqual(axes.get_ylabel(), "Activity fraction")

        inv = Inventory({"C-14": 1.0, "K-40": 2.0}, "g")
        _, axes = inv.plot(100, "ky", yunits="mass_frac")
        self.assertEqual(axes.get_ylabel(), "Mass fraction")

        inv = Inventory({"C-14": 1.0, "K-40": 2.0}, "kmol")
        _, axes = inv.plot(100, "ky", yunits="mol_frac")
        self.assertEqual(axes.get_ylabel(), "Mole fraction")

        with self.assertRaises(ValueError):
            inv.plot(100, "ky", order="invalid")
        with self.assertRaises(ValueError):
            inv.plot(100, "ky", yunits="invalid")

    def test___eq__(self) -> None:
        """
        Test Inventory equality.
        """

        inv1 = Inventory({"H-3": 10.0})
        inv2 = Inventory({"H3": 10.0})
        self.assertEqual(inv1, inv2)

        decay_data = load_dataset("icrp107_ame2020_nubase2020", load_sympy=True)
        inv2 = Inventory({"H-3": 10.0}, decay_data=decay_data)
        self.assertEqual(inv1, inv2)

        self.assertFalse(inv1 == "random object")

    def test___ne__(self) -> None:
        """
        Test Inventory inequality.
        """

        inv1 = Inventory({"H-3": 10.0})
        inv2 = Inventory({"Cs-137": 10.0})
        self.assertNotEqual(inv1, inv2)

        inv1 = Inventory({"H-3": 10.0})
        inv2 = Inventory({"H-3": 5.0})
        self.assertNotEqual(inv1, inv2)

        self.assertTrue(inv1 != "random object")

    def test___repr__(self) -> None:
        """
        Test Inventory __repr__ strings.
        """

        inv = Inventory({"H-3": 10.0}, "Bq")
        self.assertEqual(
            repr(inv),
            "Inventory activities (Bq): {'H-3': 10.0}, decay dataset: icrp107_ame2020_nubase2020",
        )


class TestInventoryHP(unittest.TestCase):
    """
    Unit tests for the inventory.py InventoryHP class.
    """

    def test_instantiation(self) -> None:
        """
        Test InventoryHP instantiation.
        """

        inv = InventoryHP({"H3": 1}, "Bq", True)
        self.assertEqual(
            inv.contents, {"H-3": Integer(242988330816) / (Integer(625) * log(2))}
        )
        self.assertEqual(inv.decay_data, DEFAULTDATA)
        self.assertEqual(inv.decay_matrices, DEFAULTDATA.sympy_data)

        temp_data = copy.deepcopy(DEFAULTDATA)
        backup_year_conv = temp_data._sympy_year_conv
        temp_data._sympy_year_conv = None
        with self.assertRaises(ValueError):
            InventoryHP({"H3": 1}, "Bq", True, temp_data)
        temp_data._sympy_year_conv = backup_year_conv
        temp_data._sympy_data = None
        with self.assertRaises(ValueError):
            InventoryHP({"H3": 1}, "Bq", True, temp_data)

        # check instantiation with a stable nuclide activity raises a ValueError
        with self.assertRaises(ValueError):
            InventoryHP({"He-3": 0.0}, "Bq", False)

    def test_numbers(self) -> None:
        """
        Test InventoryHP.numbers() method.
        """

        inv = InventoryHP({"H-3": 1}, "num")
        self.assertEqual(inv.numbers(), {"H-3": 1.0})
        inv = InventoryHP({"Tc-99m": 2.3, "I-123": 5.8})
        self.assertEqual(
            inv.numbers(), {"I-123": 399738.47946141585, "Tc-99m": 71852.27235544211}
        )

    def test_activities(self) -> None:
        """
        Test InventoryHP.activities() method.
        """

        inv = InventoryHP({"H-3": 1})
        self.assertEqual(inv.activities(), {"H-3": 1})
        inv = InventoryHP({"Tc-99m": 2.3, "I-123": 5.8})
        self.assertEqual(inv.activities(), {"I-123": 5.8, "Tc-99m": 2.3})
        inv = InventoryHP({"Tc-99m": 2300, "I-123": 5800})
        self.assertEqual(inv.activities("kBq"), {"I-123": 5.8, "Tc-99m": 2.3})

    def test_masses(self) -> None:
        """
        Test InventoryHP.masses() method.
        """

        inv = InventoryHP({"H-3": 1}, "g")
        self.assertEqual(inv.masses(), {"H-3": 1})
        inv = InventoryHP({"Tc-99m": 2.3, "I-123": 5.8})
        self.assertEqual(
            inv.masses(),
            {"I-123": 8.158243973887584e-17, "Tc-99m": 1.1800869622748503e-17},
        )
        self.assertEqual(
            inv.masses("pg"),
            {"I-123": 8.158243973887584e-05, "Tc-99m": 1.1800869622748502e-05},
        )

    def test_decay(self) -> None:
        """
        Test InventoryHP.decay() calculations.
        """

        inv = InventoryHP({"U-238": 99.274, "U-235": 0.720, "U-234": 0.005}, "Bq")
        self.assertEqual(
            inv.decay(1e9, "y").activities(),
            {
                "Ac-227": 0.26900062817405557,
                "At-218": 0.01700286863849718,
                "At-219": 2.227325201281318e-07,
                "Bi-210": 85.01434361515662,
                "Bi-211": 0.2690008442558584,
                "Bi-214": 85.01432618961894,
                "Bi-215": 2.1605054452429227e-07,
                "Fr-223": 0.003712208668802187,
                "Hg-206": 1.6152725286830195e-06,
                "Pa-231": 0.2690006198549054,
                "Pa-234": 0.13601313171698984,
                "Pa-234m": 85.00820732310412,
                "Pb-206": 0.0,
                "Pb-207": 0.0,
                "Pb-210": 85.01434361489547,
                "Pb-211": 0.26900084425585685,
                "Pb-214": 84.99734032384836,
                "Po-210": 85.01434362236536,
                "Po-211": 0.0007424423301461693,
                "Po-214": 84.99649018398776,
                "Po-215": 0.26900084425583065,
                "Po-218": 85.0143431924859,
                "Ra-223": 0.2690006282052861,
                "Ra-226": 85.0143431922866,
                "Rn-218": 1.7002868638497178e-05,
                "Rn-219": 0.26900062820528614,
                "Rn-222": 85.01434319248578,
                "Th-227": 0.26528841952452625,
                "Th-230": 85.01431274847525,
                "Th-231": 0.26898810215560653,
                "Th-234": 85.00820732310407,
                "Tl-206": 0.00011383420610068996,
                "Tl-207": 0.2682584019257157,
                "Tl-210": 0.017853008499819988,
                "U-234": 85.01287846492669,
                "U-235": 0.26898810215449415,
                "U-238": 85.00820732184867,
            },
        )

        inv.sig_fig = 0
        with self.assertRaises(ValueError):
            inv.decay(1e9, "y")

    def test_cumulative_decays(self) -> None:
        """
        Test InventoryHP.cumulative_decays() calculations.
        """

        inv = InventoryHP({"H-3": 10.0}, "num")
        self.assertEqual(inv.cumulative_decays(12.32, "y"), {"H-3": 5.0})
        self.assertEqual(inv.cumulative_decays(1e6, "y"), {"H-3": 10.0})

        inv = InventoryHP({"Sr-90": 10.0}, "num")
        self.assertEqual(inv.cumulative_decays(1e6, "y"), {"Sr-90": 10.0, "Y-90": 10.0})

        inv = InventoryHP({"Tc-99m": 2.3, "I-123": 5.8}, "num")
        self.assertEqual(
            inv.cumulative_decays(20.0, "h"),
            {
                "I-123": 3.7595407554652263,
                "Te-123m": 4.728012744186443e-07,
                "Te-123": 5.801848145646267e-18,
                "Tc-99m": 2.0704925198993647,
                "Tc-99": 1.0500063848957902e-08,
            },
        )

        inv = InventoryHP({"U-238": 99.274, "U-235": 0.720, "U-234": 0.005}, "num")
        self.assertEqual(
            inv.cumulative_decays(1e9, "y"),
            {
                "Ac-227": 0.45099937182594446,
                "Th-227": 0.4447755804754738,
                "Ra-226": 14.264656807713404,
                "Fr-223": 0.0062237913311978125,
                "Ra-223": 0.45099937179471394,
                "Rn-222": 14.264656807514218,
                "Rn-219": 0.4509993717947139,
                "At-219": 3.734274798718682e-07,
                "Po-218": 14.264656807514106,
                "At-218": 0.002852931361502821,
                "Rn-218": 2.8529313615028208e-06,
                "Bi-215": 3.6222465547570775e-07,
                "Po-215": 0.4509997340193693,
                "Pb-214": 14.261803876151633,
                "Bi-214": 14.264653954581055,
                "Po-214": 14.261661230181954,
                "Pb-211": 0.4509997340193431,
                "Bi-211": 0.45099973401934157,
                "Po-211": 0.0012447592658933826,
                "Tl-210": 0.0029955773304620116,
                "Pb-210": 14.264656385104532,
                "Bi-210": 14.264656384843379,
                "Po-210": 14.264656377634646,
                "U-238": 14.265792678151334,
                "Tl-207": 0.4497549747534447,
                "Hg-206": 2.710284713169805e-07,
                "Tl-206": 1.9100374899310035e-05,
                "U-235": 0.4510118978455059,
                "Th-234": 14.265792676895929,
                "Pa-234m": 14.265792676895886,
                "Pa-234": 0.02282526828301015,
                "U-234": 14.266121535073315,
                "Th-231": 0.4510118978443935,
                "Pa-231": 0.45099938014509455,
                "Th-230": 14.264687251524753,
            },
        )

        # Catch incorrect sig_fig or no SymPy data in decay dataset
        inv.sig_fig = 0
        with self.assertRaises(ValueError):
            inv.cumulative_decays(1e9, "y")

    @patch("matplotlib.pyplot.show")
    def test_plot(self, mock_show) -> None:
        """
        Test InventoryHP.plot() method.
        """

        inv = InventoryHP({"C-14": 1.0, "K-40": 2.0}, "mol")
        _, axes = inv.plot(
            100,
            xscale="log",
            yscale="log",
            yunits="mmol",
            display=["K40", "C14"],
        )
        self.assertEqual(axes.get_xscale(), "log")
        self.assertEqual(axes.get_yscale(), "log")
        self.assertEqual(axes.get_xlabel(), "Time (s)")
        self.assertEqual(axes.get_ylabel(), "Number of moles (mmol)")
        self.assertEqual(axes.get_xlim()[0], 0.0707945784384138)
        self.assertEqual(axes.get_ylim(), (949.999999633917, 2100.0))
        self.assertEqual(axes.get_legend_handles_labels()[-1], ["K-40", "C-14"])

    def test___repr__(self) -> None:
        """
        Test InventoryHP __repr__ strings.
        """

        inv = InventoryHP({"H-3": 10.0}, "Bq")
        self.assertEqual(
            repr(inv),
            "InventoryHP activities (Bq): {'H-3': 10.0}, decay dataset: icrp107_ame2020_nubase2020",
        )


if __name__ == "__main__":
    unittest.main()
